import pytest
from abc import abstractmethod

from plenum.test.test_node import ensureElectionsDone

from indy_common.authorize.auth_actions import split_action_id
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from plenum.common.constants import STEWARD_STRING, TRUSTEE, TRUSTEE_STRING, VALIDATOR
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import waitForViewChange
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.helper import build_auth_rule_request_json
from stp_core.loop.eventually import eventually


class NodeAuthTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.tconf = env.tconf
        self.tdir = env.tdir
        self.sdk_pool_handle = env.sdk_pool_handle
        self.client_wallet = env.sdk_wallet_client
        self.trustee_wallet = self._create_trustee(env.sdk_wallet_trustee)
        self.new_nodes = {}
        self.txnPoolNodeSet = env.txnPoolNodeSet

    def prepare(self):
        pass

    def run(self):

        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, self.trustee_wallet)

        # Step 2. Check, that we cannot send NODE txn by old way
        with pytest.raises(RequestRejectedException):
            self.send_and_check(*self.node_req_1)

        # Step 3. Check, that a new way works
        self.send_and_check(*self.node_req_for_new_rule)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.send_and_check(*self.node_req_2)

        self._demote_new_nodes()

    def result(self):
        pass

    def get_changed_auth_rule(self):
        constraint = AuthConstraint(role=TRUSTEE,
                                    sig_count=1,
                                    need_to_be_owner=False)
        params = self._generate_auth_rule_params(constraint)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1], **params
        )

    def _create_steward(self):
        return sdk_add_new_nym(self.looper, self.sdk_pool_handle,
                               self.trustee_wallet,
                               role=STEWARD_STRING)

    def _create_trustee(self, trustee_wallet):
        return sdk_add_new_nym(self.looper, self.sdk_pool_handle,
                               trustee_wallet,
                               role=TRUSTEE_STRING)

    def _add_node(self, wallet=None, services=[VALIDATOR]):
        if not wallet:
            wallet = self._create_steward()
        req, node_data, node_name = self._build_node(wallet,
                                                     self.tconf,
                                                     self.tdir,
                                                     services=services)
        self.new_nodes[wallet] = (node_data, node_name)
        return req, node_data, node_name, wallet

    def _demote_new_nodes(self):
        view_no = self.txnPoolNodeSet[0].viewNo
        for wallet, (node_data, node_name) in self.new_nodes.items():
            print("demote {}".format(node_name))
            req1, node_data1, node_name1 = self._build_node(wallet,
                                                            self.tconf,
                                                            self.tdir,
                                                            services=[],
                                                            node_name=node_name,
                                                            node_data=node_data)
            self.send_and_check(req1, wallet)
            view_no = self._wait_view_change_finish(view_no)

    def _wait_view_change_finish(self, view_no):
        view_no += 1
        waitForViewChange(looper=self.looper, txnPoolNodeSet=self.txnPoolNodeSet,
                          expectedViewNo=view_no)

        def check_not_in_view_change():
            assert all([not n.master_replica._consensus_data.waiting_for_new_view
                        for n in self.txnPoolNodeSet])

        # we may have multiple view changes since we can select the same Primary as in previous view,
        # or select a demoted node as a Primary
        self.looper.run(eventually(check_not_in_view_change, timeout=100))
        return view_no

    @abstractmethod
    def _generate_auth_rule_params(self, constraint):
        pass


class AddNodeTest(NodeAuthTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id=action_id)

    def prepare(self):
        self.node_req_1 = self.get_node_req()
        self.new_nodes.pop(self.node_req_1[1])
        self.node_req_2 = self.get_node_req()
        self.node_req_for_new_rule = self.get_node_req(self.trustee_wallet)

        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def get_node_req(self, steward_wallet=None):
        req, node_data, node_name, wallet = self._add_node(steward_wallet)
        return req, wallet

    def _generate_auth_rule_params(self, constraint):
        return dict(
            auth_action=self.action.prefix,
            auth_type=self.action.txn_type,
            field=self.action.field,
            new_value=self.action.new_value,
            constraint=constraint.as_dict
        )


class EditNodeTest(NodeAuthTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id=action_id)

    def _generate_auth_rule_params(self, constraint):
        return dict(
            auth_action=self.action.prefix,
            auth_type=self.action.txn_type,
            field=self.action.field,
            new_value=self.action.new_value,
            old_value=self.action.old_value,
            constraint=constraint.as_dict
        )


class EditNodeServicesTest(EditNodeTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.new_services = self.action.new_value

    def prepare(self):
        req, node_data, node_name, wallet = self._edit_node()
        self.node_req_1 = req, wallet

        req, node_data, node_name, wallet = self._edit_node(node_name=node_name,
                                                            node_data=node_data,
                                                            wallet=self.trustee_wallet)
        self.node_req_for_new_rule = req, self.trustee_wallet

        req, node_data, node_name, wallet = self._edit_node()
        self.node_req_2 = req, wallet

        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def run(self):
        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, self.trustee_wallet)

        # Step 2. Check, that we cannot send NODE txn by old way
        with pytest.raises(RequestRejectedException):
            self.send_and_check(*self.node_req_1)

        view_no = self.txnPoolNodeSet[0].viewNo

        # Step 3. Check, that a new way works
        self.send_and_check(*self.node_req_for_new_rule)

        view_no = self._wait_view_change_finish(view_no)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.send_and_check(*self.node_req_2)

        view_no = self._wait_view_change_finish(view_no)

        self._demote_new_nodes()

    def _edit_node(self, wallet=None, services=[VALIDATOR], node_name=None, node_data=None):
        if not (node_name and node_data and wallet):
            req, node_data, node_name, wallet = self._add_node()
            self.send_and_check(req, wallet)
        req, node_data, node_name = self._build_node(wallet,
                                                     self.tconf,
                                                     self.tdir,
                                                     services=services,
                                                     node_name=node_name,
                                                     node_data=node_data)
        return req, node_data, node_name, wallet


class AddNewNodeTest(AddNodeTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def run(self):
        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, self.trustee_wallet)

        # Step 2. Check, that we cannot send NODE txn by old way
        with pytest.raises(RequestRejectedException):
            self.send_and_check(*self.node_req_1)

        prev_view_no = self.txnPoolNodeSet[0].viewNo
        new_node_count = len(self.new_nodes)

        # Step 3. Check, that a new way works
        self.send_and_check(*self.node_req_for_new_rule)

        prev_view_no = self._wait_view_change_finish(prev_view_no)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.send_and_check(*self.node_req_2)

        prev_view_no = self._wait_view_change_finish(prev_view_no)

        self._demote_new_nodes()


class AddNewNodeEmptyServiceTest(AddNodeTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_node(self, wallet=None, services=[]):
        return super()._add_node(wallet, services)

    def _demote_new_nodes(self):
        # Pass demote_new_nodes because in this test
        # all nodes create with service=[]
        pass


class DemoteNodeTest(EditNodeServicesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _edit_node(self, wallet=None, services=[], node_name=None, node_data=None):
        return super()._edit_node(wallet, services, node_name, node_data)

    def _demote_new_nodes(self):
        pass


class PromoteNodeTest(EditNodeServicesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_node(self, wallet=None, services=[]):
        return super()._add_node(wallet, services)



import json

import pytest

from indy_common.constants import JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, RICH_SCHEMA, RICH_SCHEMA_ENCODING, \
    RICH_SCHEMA_MAPPING, RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE, RS_MAPPING_TYPE_VALUE, \
    RS_ENCODING_TYPE_VALUE, RS_SCHEMA_TYPE_VALUE, RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE
from indy_node.test.api.helper import validate_write_reply, validate_rich_schema_txn, sdk_build_rich_schema_request
from indy_node.test.helper import rich_schemas_enabled_scope
from indy_node.test.rich_schema.templates import RICH_SCHEMA_EX1, W3C_BASE_CONTEXT, RICH_SCHEMA_ENCODING_EX1, \
    RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_CRED_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1
from plenum.common.util import randomString
from plenum.test.helper import sdk_get_reply, sdk_sign_and_submit_req


@pytest.fixture(scope="module")
def tconf(tconf):
    with rich_schemas_enabled_scope(tconf):
        yield tconf


# The order of creation is essential as some rich schema object reference others by ID
# Encoding's id must be equal to the one used in RICH_SCHEMA_MAPPING_EX1

@pytest.mark.parametrize('txn_type, rs_type, content, rs_id',
                         [(JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, W3C_BASE_CONTEXT, randomString()),
                          (RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, RICH_SCHEMA_EX1, RICH_SCHEMA_EX1['@id']),
                          (RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA_ENCODING_EX1,
                           "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD"),
                          (RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_MAPPING_EX1,
                           RICH_SCHEMA_MAPPING_EX1['@id']),
                          (RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA_CRED_DEF_EX1, randomString()),
                          (RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE, RICH_SCHEMA_PRES_DEF_EX1,
                           RICH_SCHEMA_PRES_DEF_EX1['@id'])])
def test_rich_schema_object_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward,
                                           txn_type, rs_type, content, rs_id):
    request = sdk_build_rich_schema_request(looper, sdk_wallet_steward,
                                            txn_type=txn_type, rs_id=rs_id, rs_name=randomString(),
                                            rs_version='1.0', rs_type=rs_type,
                                            rs_content=json.dumps(content))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request))[1]
    validate_write_reply(reply)
    validate_rich_schema_txn(reply['result']['txn'], txn_type)



from datetime import datetime, timedelta

import dateutil.tz
import pytest

from stp_core.loop.eventually import eventually
from plenum.common.constants import VERSION
from plenum.test.helper import randomText

from indy_common.constants import START, FORCE, APP_NAME
from indy_common.version import src_version_cls
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion
from indy_node.server.upgrader import Upgrader

from indy_node.test import waits
from indy_node.test.upgrade.helper import bumpedVersion, \
    checkUpgradeScheduled, bumpVersion, sdk_ensure_upgrade_sent


@pytest.fixture(scope='module')
def nodeIds(nodeSet):
    return nodeSet[0].poolManager.nodeIds


EXT_PKT_NAME = 'SomeTopLevelPkt'
EXT_PKT_VERSION = '7.88.999'


# TODO review and improve: which contexts are goinig to be satisfied
# by the following mocks
def patch_packet_mgr_output(monkeypatch, pkg_name, pkg_version, new_version):
    if pkg_name != APP_NAME:
        node_package = (APP_NAME, '0.0.1')
        EXT_TOP_PKT_DEPS = [("aa", "1.1.1"), ("bb", "2.2.2")]
        PACKAGE_MNG_EXT_PTK_OUTPUT = (
            "Package: {}\nVersion: {}\n"
            "Depends: {}, {} (= {}), {} (= {})\n"
            .format(pkg_name, pkg_version, APP_NAME,
                    *EXT_TOP_PKT_DEPS[0], *EXT_TOP_PKT_DEPS[1])
        )
        top_level_package = (pkg_name, pkg_version)
        plenum_package = ('indy-plenum', '0.0.3')
        top_level_package_with_version = '{}={}'.format(*top_level_package)
        top_level_package_dep1_with_version = '{}={}'.format(*EXT_TOP_PKT_DEPS[0])
        top_level_package_dep2_with_version = '{}={}'.format(*EXT_TOP_PKT_DEPS[1])
        node_package_with_version = '{}={}'.format(*node_package)
        plenum_package_with_version = '{}={}'.format(*plenum_package)
        mock_info = {
            top_level_package_with_version: "{}{} (= {}) {} (= {}), {} (= {})".format(
                randomText(100), *node_package, *EXT_TOP_PKT_DEPS[0], *EXT_TOP_PKT_DEPS[1]),
            node_package_with_version: '{}{} (= {}){}{}'.format(
                randomText(100), *plenum_package, randomText(100), randomText(100)),
            plenum_package_with_version: '{}'.format(randomText(100)),
            top_level_package_dep1_with_version: '{}{} (= {})'.format(randomText(100), *plenum_package),
            top_level_package_dep2_with_version: '{}{} (= {})'.format(randomText(100), *node_package)
        }

        def mock_get_info_from_package_manager(package):
            if package.startswith(pkg_name):
                return mock_info.get(top_level_package_with_version, "")
            return mock_info.get(package, "")

        #monkeypatch.setattr(NodeControlUtil, '_get_info_from_package_manager',
        #                    lambda x: mock_get_info_from_package_manager(x))
        monkeypatch.setattr(NodeControlUtil, '_get_curr_info', lambda *x: PACKAGE_MNG_EXT_PTK_OUTPUT)
    else:
        #monkeypatch.setattr(
        #    NodeControlUtil, '_get_info_from_package_manager',
        #    lambda package: "Package: {}\nVersion: {}\n".format(APP_NAME, pkg_version) if package == APP_NAME else ""
        #)
        monkeypatch.setattr(
            NodeControlUtil, '_get_curr_info',
            lambda *x: "Package: {}\nVersion: {}\n".format(APP_NAME, pkg_version)
        )

    monkeypatch.setattr(NodeControlUtil, 'update_package_cache', lambda *x: None)
    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: DebianVersion(
            new_version, upstream_cls=src_version_cls(pkg_name))
    )


@pytest.fixture(scope='function', params=[
    (EXT_PKT_NAME, EXT_PKT_VERSION),
    (APP_NAME, Upgrader.get_src_version(APP_NAME).release)
])
def pckg(request):
    return request.param


@pytest.fixture(scope='function')
def validUpgrade(nodeIds, tconf, monkeypatch, pckg):
    schedule = {}
    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    startAt = unow + timedelta(seconds=100)
    acceptableDiff = tconf.MinSepBetweenNodeUpgrades + 1
    for i in nodeIds:
        schedule[i] = datetime.isoformat(startAt)
        startAt = startAt + timedelta(seconds=acceptableDiff + 3)

    new_version = bumpedVersion(pckg[1])
    patch_packet_mgr_output(monkeypatch, pckg[0], pckg[1], new_version)

    return dict(name='upgrade-{}'.format(randomText(3)), version=new_version,
                action=START, schedule=schedule, timeout=1, package=pckg[0],
                sha256='db34a72a90d026dae49c3b3f0436c8d3963476c77468ad955845a1ccf7b03f55')


@pytest.fixture(scope='function')
def validUpgradeExpForceFalse(validUpgrade):
    nup = validUpgrade.copy()
    nup.update({FORCE: False})
    nup.update({VERSION: bumpVersion(validUpgrade[VERSION])})
    return nup


@pytest.fixture(scope='function')
def validUpgradeExpForceTrue(validUpgradeExpForceFalse):
    nup = validUpgradeExpForceFalse.copy()
    nup.update({FORCE: True})
    nup.update({VERSION: bumpVersion(validUpgradeExpForceFalse[VERSION])})
    return nup


@pytest.fixture(scope='function')
def validUpgradeSent(looper, nodeSet, tdir, sdk_pool_handle, sdk_wallet_trustee,
                     validUpgrade):
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, validUpgrade)


@pytest.fixture(scope='function')
def validUpgradeSentExpForceFalse(
        looper,
        nodeSet,
        tdir,
        sdk_pool_handle,
        sdk_wallet_trustee,
        validUpgradeExpForceFalse):
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, validUpgradeExpForceFalse)


@pytest.fixture(scope='function')
def validUpgradeSentExpForceTrue(looper, nodeSet, tdir, sdk_pool_handle,
                                 sdk_wallet_trustee, validUpgradeExpForceTrue):
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, validUpgradeExpForceTrue)


@pytest.fixture(scope='function')
def upgradeScheduled(validUpgradeSent, looper, nodeSet, validUpgrade):
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            validUpgrade[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))


@pytest.fixture(scope='function')
def upgradeScheduledExpForceFalse(validUpgradeSentExpForceFalse, looper,
                                  nodeSet, validUpgradeExpForceFalse):
    looper.run(eventually(checkUpgradeScheduled, nodeSet,
                          validUpgradeExpForceFalse[VERSION], retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))


@pytest.fixture(scope='function')
def upgradeScheduledExpForceTrue(validUpgradeSentExpForceTrue, looper, nodeSet,
                                 validUpgradeExpForceTrue):
    looper.run(eventually(checkUpgradeScheduled, nodeSet,
                          validUpgradeExpForceTrue[VERSION], retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))


@pytest.fixture(scope='function')
def invalidUpgrade(nodeIds, tconf, validUpgrade):
    nup = validUpgrade.copy()
    schedule = {}
    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    startAt = unow + timedelta(seconds=60)
    acceptableDiff = tconf.MinSepBetweenNodeUpgrades + 1
    for i in nodeIds:
        schedule[i] = datetime.isoformat(startAt)
        startAt = startAt + timedelta(seconds=acceptableDiff - 3)
    nup.update(dict(name='upgrade-14', version=bumpedVersion(), action=START,
                    schedule=schedule,
                    sha256='46c715a90b1067142d548cb1f1405b0486b32b1a27d418ef3a52bd976e9fae50',
                    timeout=10))
    return nup



from indy_common.constants import ID, GET_REVOC_REG_DEF

from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler


class GetRevocRegDefHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, GET_REVOC_REG_DEF, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        state_path = request.operation.get(ID, None)
        assert state_path
        try:
            keys, last_seq_no, last_update_time, proof = self.lookup(state_path, is_committed=True, with_proof=True)
        except KeyError:
            keys, last_seq_no, last_update_time, proof = None, None, None, None
        result = self.make_result(request=request,
                                  data=keys,
                                  last_seq_no=last_seq_no,
                                  update_time=last_update_time,
                                  proof=proof)
        return result



from types import MethodType

import logging
import pytest

from indy_node.server.node import Node
from plenum.common.batched import Batched
from plenum.server.consensus.ordering_service import OrderingService
from plenum.test.delayers import cDelay, lsDelay
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.test_node import getNonPrimaryReplicas
from stp_core.common.log import getlogger
from stp_core.loop.eventually import eventually

logger = getlogger()
whitelist = ["is not connected - message will not be sent immediately."]


@pytest.fixture(scope="module")
def tconf(tconf):
    oldMax3PCBatchSize = tconf.Max3PCBatchSize
    tconf.Max3PCBatchSize = 1

    yield tconf

    tconf.Max3PCBatchSize = oldMax3PCBatchSize


@pytest.fixture(scope="module")
def disable_transport_batching():
    original_should_batch = Batched._should_batch
    Batched._should_batch = lambda self, msgs: False

    yield

    Batched._should_batch = original_should_batch


# Enable blowing up for log messages with WARNING and higher severity levels
concerningLogLevels = [logging.WARNING,
                       logging.ERROR,
                       logging.CRITICAL]


def test_batch_rejected_on_catchup_start_can_be_ordered_before_ledgers_sync(
        looper,
        tdirWithPoolTxns,
        tdirWithDomainTxns,
        nodeSet,
        sdk_pool_handle,
        sdk_wallet_endorser,
        allPluginsPath,
        tconf,
        disable_transport_batching):
    """
    Verifies that a batch rejected due to catch-up start can be successfully
    ordered later before ledgers synchronization without any
    warnings.

    In the test we perform stashing / unstashing messages and patching /
    unpatching methods to ensure the desired order of events for the following
    scenario:

    1. Start a pool of 4 nodes. One node with a master replica not being
    a primary suffers from a slow network connection. We name this node slow.
    2. Start a client.
    3. Send a write request from the client to the pool.
    4. All the 3PC messages for this request are sent, received and handled by
    all the nodes except that COMMIT messages are not received by the slow node.
    5. The other nodes reach a consensus and order the batch.
    6. Initiate a catch-up of the slow node by calling its start_catchup method.
    7. The slow node rejects the unordered batch.
    8. The slow node sends MESSAGE_REQUEST for LEDGER_STATUS to the other nodes.
    9. The slow node receives the COMMIT messages from the other nodes.
    10. The slow node gets ORDER messages from its replicas and stashes them
    since it is not in participating mode.
    11. The slow node receives LEDGER_STATUS messages from the other nodes.
    12. Prior to ledgers synchronization the slow node processes the stashed
    Commits messages and orders batch.
    13. The slow node synchronizes its ledgers and completes the catch-up.
    """

    non_primary_replicas_of_master = getNonPrimaryReplicas(nodeSet, 0)
    slow_node = non_primary_replicas_of_master[0].node
    slow_os = slow_node.master_replica._ordering_service

    slow_node.nodeIbStasher.delay(cDelay(300))
    slow_node.start_catchup = MethodType(patched_start_catchup, slow_node)

    no_more_catchups_needed_call_times_before = \
        slow_node.spylog.count(Node.no_more_catchups_needed.__name__)
    on_batch_rejected_call_times_before = \
        slow_os.spylog.count(OrderingService.post_batch_rejection.__name__)
    on_batch_created_call_times_before = \
        slow_os.spylog.count(OrderingService.post_batch_creation.__name__)
    process_ordered_call_times_before = \
        slow_node.spylog.count(Node.processOrdered.__name__)

    send_random_requests(looper, sdk_pool_handle, sdk_wallet_endorser, 1)

    slow_node.start_catchup()

    def check_catchup_done():
        assert slow_node.spylog.count(Node.no_more_catchups_needed.__name__) > \
               no_more_catchups_needed_call_times_before

    looper.run(eventually(check_catchup_done, retryWait=1, timeout=10))

    on_batch_rejected_call_times_after = \
        slow_os.spylog.count(OrderingService.post_batch_rejection.__name__)
    on_batch_created_call_times_after = \
        slow_os.spylog.count(OrderingService.post_batch_creation.__name__)
    process_ordered_call_times_after = \
        slow_node.spylog.count(Node.processOrdered.__name__)

    assert on_batch_rejected_call_times_after \
           - on_batch_rejected_call_times_before == 1

    assert on_batch_created_call_times_after \
           - on_batch_created_call_times_before == 1

    assert process_ordered_call_times_after \
           - process_ordered_call_times_before == 0  # one per replica
    last_2_process_ordered_results = \
        [call.result for call
         in slow_node.spylog.getAll(Node.processOrdered.__name__)[-2:]]
    assert True in last_2_process_ordered_results  # True for master replica


def patched_start_catchup(self, *args, **kwargs):
    Node.start_catchup(self, *args, **kwargs)

    self.nodeIbStasher.reset_delays_and_process_delayeds()
    self.nodeIbStasher.delay(lsDelay(300))

    self.try_processing_ordered = \
        MethodType(patched_try_processing_ordered, self)
    self.start_catchup = MethodType(Node.start_catchup, self)


def patched_try_processing_ordered(self, msg):
    Node.try_processing_ordered(self, msg)

    if msg.instId == 0:
        self.nodeIbStasher.reset_delays_and_process_delayeds()

        self.try_processing_ordered = \
            MethodType(Node.try_processing_ordered, self)


def send_random_requests(looper, sdk_pool_handle, sdk_wallet_endorser, count: int):
    logger.info('{} random requests will be sent'.format(count))
    for i in range(count):
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_endorser)



import copy
import json
import random

import pytest

from indy_common.constants import REVOC_TYPE, ACCUM, PREV_ACCUM, REVOC_REG_DEF_ID, REVOC_REG_ENTRY, VALUE, ISSUED, \
    REVOKED
from plenum.common.constants import REQNACK, REJECT, TXN_TYPE, GENERAL_LIMIT_SIZE
from plenum.common.util import randomString
from plenum.test.helper import sdk_send_signed_requests, sdk_get_replies, sdk_sign_request_from_dict


@pytest.fixture(scope="module", params=['lt', 'eq', 'gt'])
def _lt_eq_gt(request):
    return request.param


@pytest.fixture(scope="module")
def _entry_skeleton():
    return {
        REVOC_REG_DEF_ID: randomString(10),
        TXN_TYPE: REVOC_REG_ENTRY,
        REVOC_TYPE: randomString(10),
        VALUE: {
            PREV_ACCUM: randomString(10),
            ACCUM: randomString(10),
            ISSUED: [],
            REVOKED: [],
        }
    }


@pytest.fixture(scope="module", params=[REVOC_TYPE])
def _res_field_size(request, _lt_eq_gt):
    _field = request.param
    _expected = REQNACK if _lt_eq_gt == 'gt' else REJECT
    _valid_size = GENERAL_LIMIT_SIZE

    if _lt_eq_gt == 'lt':
        return _expected, _field, random.randint(0, _valid_size - 1)
    if _lt_eq_gt == 'eq':
        return _expected, _field, _valid_size
    return _expected, _field, random.randint(_valid_size + 1, 2 * _valid_size)


@pytest.fixture(scope="module")
def revoc_entry(looper,
                sdk_wallet_steward,
                _entry_skeleton,
                _lt_eq_gt,
                _res_field_size):
    _expected, _field, _size = _res_field_size
    _req = copy.deepcopy(_entry_skeleton)
    _value = randomString(_size)

    _req[_field] = _value

    return _expected, sdk_sign_request_from_dict(looper, sdk_wallet_steward, _req)


def test_revoc_entry_static_validation_on_size(revoc_entry,
                                               looper,
                                               txnPoolNodeSet,
                                               sdk_pool_handle,
                                               sdk_wallet_steward):
    _expected, _req = revoc_entry
    results = sdk_send_signed_requests(sdk_pool_handle, [json.dumps(_req)])
    _reply = sdk_get_replies(looper, results)[0][1]
    assert _expected == _reply['op']



import copy

import pytest

from common.serializers.serialization import ledger_txn_serializer, domain_state_serializer
from indy_common.authorize.auth_constraints import AuthConstraint, AuthConstraintOr, AuthConstraintAnd, \
    ConstraintsSerializer, ROLE, CONSTRAINT_ID, ConstraintsEnum, NEED_TO_BE_OWNER, SIG_COUNT, METADATA, \
    ConstraintCreator, AUTH_CONSTRAINTS, AuthConstraintForbidden
from plenum.common.constants import TRUSTEE, STEWARD


def test_str_any_7_off_ledger_sig():
    constraint = AuthConstraint(role='*',
                                sig_count=7,
                                off_ledger_signature=True)
    assert str(constraint) == '7 signatures of any role (off-ledger included) are required'


def test_str_any_1_off_ledger_sig():
    constraint = AuthConstraint(role='*',
                                sig_count=1,
                                off_ledger_signature=True)
    assert str(constraint) == '1 signature of any role (off-ledger included) is required'


def test_str_not_any_7_sig_owner():
    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=7,
                                need_to_be_owner=True)
    assert str(constraint) == '7 TRUSTEE signatures are required and needs to be owner'


def test_str_not_any_7_sig_not_owner():
    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=7,
                                need_to_be_owner=False)
    assert str(constraint) == '7 TRUSTEE signatures are required'


def test_str_not_any_1_sig_not_owner():
    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=1,
                                need_to_be_owner=False)
    assert str(constraint) == '1 TRUSTEE signature is required'


def test_str_not_any_1_sig_owner():
    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=1,
                                need_to_be_owner=True)
    assert str(constraint) == '1 TRUSTEE signature is required and needs to be owner'


def test_str_any_1_sig_owner():
    constraint = AuthConstraint(role="*",
                                sig_count=1,
                                need_to_be_owner=True)
    assert str(constraint) == '1 signature of any role is required and needs to be owner'


def test_str_any_1_sig_not_owner():
    constraint = AuthConstraint(role='*',
                                sig_count=1,
                                need_to_be_owner=False)
    assert str(constraint) == '1 signature of any role is required ALL'


def test_str_any_several_sig_not_owner():
    constraint = AuthConstraint(role='*',
                                sig_count=7,
                                need_to_be_owner=False)
    assert str(constraint) == '7 signatures of any role are required'


def test_str_any_several_sig_owner():
    constraint = AuthConstraint(role='*',
                                sig_count=7,
                                need_to_be_owner=True)
    assert str(constraint) == '7 signatures of any role are required and needs to be owner'


def test_str_for_auth_constraint_or():
    constraint = AuthConstraintOr([AuthConstraint(role=TRUSTEE,
                                                  sig_count=1,
                                                  need_to_be_owner=True),
                                   AuthConstraint(role=STEWARD,
                                                  sig_count=1,
                                                  need_to_be_owner=True)])
    assert str(constraint) == '1 TRUSTEE signature is required and needs to be owner ' \
                              'OR ' \
                              '1 STEWARD signature is required and needs to be owner'


def test_str_for_auth_constraint_and():
    constraint = AuthConstraintAnd([AuthConstraint(role=TRUSTEE,
                                                   sig_count=1,
                                                   need_to_be_owner=True),
                                    AuthConstraint(role=STEWARD,
                                                   sig_count=1,
                                                   need_to_be_owner=True)])
    assert str(constraint) == '1 TRUSTEE signature is required and needs to be owner ' \
                              'AND ' \
                              '1 STEWARD signature is required and needs to be owner'


def test_str_for_auth_constraint_forbidden():
    constraint = AuthConstraintForbidden()
    assert str(constraint) == 'The action is forbidden'


@pytest.fixture
def constraints():
    role_constraints = [AuthConstraint(role=TRUSTEE,
                                       sig_count=1,
                                       need_to_be_owner=True),
                        AuthConstraint(role=STEWARD,
                                       sig_count=1,
                                       need_to_be_owner=True)]
    and_constraint = AuthConstraintAnd(role_constraints)
    or_constraint = AuthConstraintOr(role_constraints)
    forbidden_constraint = AuthConstraintForbidden()
    return role_constraints[0], and_constraint, or_constraint, forbidden_constraint


def test_check_equal(constraints):
    same_role, same_and, same_or, same_forbidden = constraints
    """
    The same constraints as from fixture but should have not the same id
    """
    role_constraints = [AuthConstraint(role=TRUSTEE,
                                       sig_count=1,
                                       need_to_be_owner=True),
                        AuthConstraint(role=STEWARD,
                                       sig_count=1,
                                       need_to_be_owner=True)]
    and_constraint = AuthConstraintAnd(role_constraints)
    or_constraint = AuthConstraintOr(role_constraints)
    forbidden_constraint = AuthConstraintForbidden()

    assert id(same_role) != id(role_constraints[0])
    assert id(same_or) != id(or_constraint)
    assert id(same_and) != id(and_constraint)
    assert id(same_forbidden) != id(forbidden_constraint)
    assert same_role == role_constraints[0]
    assert same_or == or_constraint
    assert same_and == and_constraint
    assert same_forbidden == forbidden_constraint


def test_constraint_serialization(constraints):
    constraint_serializer = ConstraintsSerializer(domain_state_serializer)
    for constraint in constraints:
        serialized = constraint_serializer.serialize(constraint)
        deserialized = constraint_serializer.deserialize(serialized)
        assert constraint == deserialized
        assert constraint_serializer.serialize(constraint) == \
            constraint_serializer.serialize(deserialized)


@pytest.fixture
def constraints_as_dict():
    trustee_role_as_dict = {
        CONSTRAINT_ID: ConstraintsEnum.ROLE_CONSTRAINT_ID,
        ROLE: TRUSTEE,
        NEED_TO_BE_OWNER: True,
        SIG_COUNT: 1,
        METADATA: {}
    }
    steward_role_as_dict = {
        CONSTRAINT_ID: ConstraintsEnum.ROLE_CONSTRAINT_ID,
        ROLE: STEWARD,
        NEED_TO_BE_OWNER: True,
        SIG_COUNT: 1,
        METADATA: {}
    }
    and_as_dict = {
        CONSTRAINT_ID: ConstraintsEnum.AND_CONSTRAINT_ID,
        AUTH_CONSTRAINTS: [
            copy.copy(trustee_role_as_dict),
            copy.copy(steward_role_as_dict),
        ]
    }
    or_as_dict = {
        CONSTRAINT_ID: ConstraintsEnum.OR_CONSTRAINT_ID,
        AUTH_CONSTRAINTS: [
            copy.copy(trustee_role_as_dict),
            copy.copy(steward_role_as_dict)
        ]
    }
    forbidden_as_dict = {
        CONSTRAINT_ID: ConstraintsEnum.FORBIDDEN_CONSTRAINT_ID
    }
    return trustee_role_as_dict, and_as_dict, or_as_dict, forbidden_as_dict


# def test_error_on_deserialize():
#     constraint_serializer = ConstraintsSerializer(domain_state_serializer)
#     constraint_serializer.deserialize(b"some_string")


def test_constraint_creator(constraints_as_dict,
                            constraints):
    from_creator = [ConstraintCreator.create_constraint(c) for c in constraints_as_dict]
    assert all(map(lambda a, b: a == b, constraints, from_creator))



import dateutil
import dateutil.tz
import pytest

from datetime import datetime, timedelta
from copy import deepcopy
from indy_common.constants import START
from indy_node.test.upgrade.conftest import patch_packet_mgr_output, EXT_PKT_NAME, EXT_PKT_VERSION

from indy_node.server.upgrader import Upgrader
from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent, bumpedVersion
from plenum.test.helper import randomText

from stp_core.common.log import getlogger

delta = 2
logger = getlogger()


@pytest.fixture(scope='function')
def pckg():
    return (EXT_PKT_NAME, EXT_PKT_VERSION)


@pytest.fixture(scope="module")
def tconf(tconf):
    old_delta = tconf.MinSepBetweenNodeUpgrades
    tconf.MinSepBetweenNodeUpgrades = delta
    yield tconf
    tconf.MinSepBetweenNodeUpgrades = old_delta


@pytest.fixture(scope='function')
def skip_functions():
    # Do this to prevent exceptions because of node_control_tool absence
    old_action_failed = deepcopy(Upgrader._action_failed)

    Upgrader._action_failed = lambda *args, **kwargs: 1
    yield
    Upgrader._action_failed = old_action_failed


def test_node_doesnt_retry_upgrade(looper, nodeSet, validUpgrade, nodeIds,
                                   sdk_pool_handle, sdk_wallet_trustee, tconf,
                                   skip_functions):
    schedule = {}
    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    startAt = unow + timedelta(seconds=delta)
    for i in nodeIds:
        schedule[i] = datetime.isoformat(startAt)
        startAt = startAt + timedelta(seconds=delta)
    validUpgrade['schedule'] = schedule

    # Emulating connection problems
    for node in nodeSet:
        node.upgrader.retry_limit = 0

    # Send upgrade
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, validUpgrade)
    looper.runFor(len(nodeIds) * delta)

    # Every node, including bad_node, tried to upgrade only once
    for node in nodeSet:
        assert node.upgrader.spylog.count(Upgrader.processLedger.__name__) == 1



from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.constants import POOL_UPGRADE, ACTION


def test_pool_upgrade_start(write_request_validation, is_owner, req):
    authorized = req.identifier == "trustee_identifier"
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=POOL_UPGRADE,
                                                                 field=ACTION,
                                                                 value='start',
                                                                 is_owner=is_owner)])


def test_pool_upgrade_cancel(write_request_validation, is_owner, req):
    authorized = req.identifier == "trustee_identifier"
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=POOL_UPGRADE,
                                                                  field=ACTION,
                                                                  old_value='start',
                                                                  new_value='cancel',
                                                                  is_owner=is_owner)])


def test_pool_upgrade_cancel_wrong_new_value(write_request_validation, is_owner, req):
    assert not write_request_validation(req,
                                        [AuthActionEdit(txn_type=POOL_UPGRADE,
                                                        field=ACTION,
                                                        old_value='start',
                                                        new_value='aa_cancel_aa',
                                                        is_owner=is_owner)])


def test_pool_upgrade_cancel_wrong_old_value(write_request_validation, is_owner, req):
    assert not write_request_validation(req,
                                        [AuthActionEdit(txn_type=POOL_UPGRADE,
                                                        field=ACTION,
                                                        old_value='not_start',
                                                        new_value='cancel',
                                                        is_owner=is_owner)])



from datetime import datetime, timedelta

import dateutil
import pytest

from indy_common.config_helper import NodeConfigHelper
from indy_node.test.conftest import sdk_node_theta_added
from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent
from plenum.common.constants import TXN_TYPE, DATA, VERSION, CURRENT_PROTOCOL_VERSION

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.authorize.auth_map import add_new_identity_owner, txn_author_agreement, \
    txn_author_agreement_aml
from indy_common.constants import NODE_UPGRADE, ACTION, COMPLETE, AUTH_RULE, START
from indy_node.test.helper import sdk_send_and_check_auth_rule_request, TestNode
from plenum.common.request import Request
from plenum.common.types import f
from plenum.test.node_catchup.helper import waitNodeDataEquality

from plenum.test.helper import assertEquality
from stp_core.loop.eventually import eventually


@pytest.fixture(scope="module")
def tconf(tconf):
    old_version_matching = tconf.INDY_VERSION_MATCHING
    tconf.INDY_VERSION_MATCHING = {"1.1.58": "1.9.1"}
    yield tconf
    tconf.INDY_VERSION_MATCHING = old_version_matching


@pytest.fixture(scope='module')
def auth_keys(nodeSet, tconf):
    return [add_new_identity_owner, txn_author_agreement, txn_author_agreement_aml]


@pytest.fixture(scope='module')
def valid_upgrade(nodeSet, tconf):
    schedule = {}
    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    startAt = unow + timedelta(seconds=30000000)
    acceptableDiff = tconf.MinSepBetweenNodeUpgrades + 1
    for n in nodeSet[0].poolManager.nodeIds:
        schedule[n] = datetime.isoformat(startAt)
        startAt = startAt + timedelta(seconds=acceptableDiff + 3)

    return dict(name='upgrade', version="10000.10.10",
                action=START, schedule=schedule, timeout=1,
                package=None,
                sha256='db34a72a90d026dae49c3b3f0436c8d3963476c77468ad955845a1ccf7b03f55')


def send_node_upgrades(nodes, version, looper, count=None):
    if count is None:
        count = len(nodes)
    last_ordered = nodes[0].master_last_ordered_3PC[1]
    for node in nodes[:count]:
        op = {
            TXN_TYPE: NODE_UPGRADE,
            DATA: {
                ACTION: COMPLETE,
                VERSION: version
            }
        }
        op[f.SIG.nm] = node.wallet.signMsg(op[DATA])

        request = node.wallet.signRequest(
            Request(operation=op, protocolVersion=CURRENT_PROTOCOL_VERSION))

        node.startedProcessingReq(request.key, node.nodestack.name)
        node.send(request)
    looper.run(eventually(lambda: assertEquality(nodes[0].master_last_ordered_3PC[1],
                                                 last_ordered + count)))


def send_auth_rule(looper,
                   sdk_pool_handle,
                   sdk_wallet_trustee,
                   original_action):
    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX, auth_type=original_action.txn_type,
                                         field=original_action.field, new_value=original_action.value,
                                         old_value=None, constraint=AuthConstraint(role="*",
                                                                                   sig_count=1).as_dict)


def test_state_recovering_for_auth_rule(nodeSet, looper, sdk_pool_handle, sdk_wallet_steward,
                                        sdk_wallet_trustee,
                                        tdir, tconf,
                                        allPluginsPath,
                                        valid_upgrade,
                                        auth_keys,
                                        monkeypatch):
    version1 = "1.1.58"
    version2 = "1.1.88"

    # send POOL_UPGRADE to write in a ledger
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                            valid_upgrade)
    send_node_upgrades(nodeSet, version1, looper, count=1)
    send_auth_rule(looper,
                   sdk_pool_handle,
                   sdk_wallet_trustee,
                   auth_keys[0])
    send_node_upgrades(nodeSet, version1, looper)
    for n in nodeSet:
        handler = n.write_manager.request_handlers.get(AUTH_RULE)[0]
        handler_for_1_9_1 = n.write_manager._request_handlers_with_version.get((AUTH_RULE, "1.9.1"))[0]
        monkeypatch.setattr(handler, '_update_auth_constraint',
                            handler_for_1_9_1._update_auth_constraint)
    send_auth_rule(looper,
                   sdk_pool_handle,
                   sdk_wallet_trustee,
                   auth_keys[1])
    monkeypatch.undo()
    send_node_upgrades(nodeSet, version2, looper)
    send_auth_rule(looper,
                   sdk_pool_handle,
                   sdk_wallet_trustee,
                   auth_keys[2])

    new_steward_wallet, new_node = sdk_node_theta_added(looper,
                                                        nodeSet,
                                                        tdir,
                                                        tconf,
                                                        sdk_pool_handle,
                                                        sdk_wallet_trustee,
                                                        allPluginsPath,
                                                        node_config_helper_class=NodeConfigHelper,
                                                        testNodeClass=TestNode)
    waitNodeDataEquality(looper, new_node, *nodeSet[:-1], exclude_from_check=['check_last_ordered_3pc_backup'])



from indy_common.authorize import auth_map


def test_auth_map_node():
    node_rules = [(auth_map.adding_new_node, "0--ADD--services--*--['VALIDATOR']"),
                  (auth_map.adding_new_node_with_empty_services, "0--ADD--services--*--[]"),
                  (auth_map.demote_node, "0--EDIT--services--['VALIDATOR']--[]"),
                  (auth_map.promote_node, "0--EDIT--services--[]--['VALIDATOR']"),
                  (auth_map.change_node_ip, '0--EDIT--node_ip--*--*'),
                  (auth_map.change_node_port, '0--EDIT--node_port--*--*'),
                  (auth_map.change_client_ip, '0--EDIT--client_ip--*--*'),
                  (auth_map.change_client_port, '0--EDIT--client_port--*--*'),
                  (auth_map.change_bls_key, '0--EDIT--blskey--*--*')]

    for (rule, rule_str) in node_rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_nym():
    nym_rules = [(auth_map.add_new_trustee, "1--ADD--role--*--0"),
                 (auth_map.add_new_steward, "1--ADD--role--*--2"),
                 (auth_map.add_new_endorser, "1--ADD--role--*--101"),
                 (auth_map.add_new_network_monitor, "1--ADD--role--*--201"),
                 (auth_map.add_new_identity_owner, '1--ADD--role--*--'),
                 (auth_map.key_rotation, '1--EDIT--verkey--*--*')]

    for (rule, rule_str) in nym_rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_txn_author_agreement():
    rules = [(auth_map.txn_author_agreement, "4--ADD--*--*--*"), ]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_txn_author_agreement_aml():
    rules = [(auth_map.txn_author_agreement_aml, "5--ADD--*--*--*"), ]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_attrib():
    rules = [(auth_map.add_attrib, "100--ADD--*--*--*"),
             (auth_map.edit_attrib, "100--EDIT--*--*--*")]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_schema():
    rules = [(auth_map.add_schema, "101--ADD--*--*--*")]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_schema_for_omitted():
    rules = [(auth_map.edit_schema, "101--EDIT--*--*--*")]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_claim_def():
    rules = [(auth_map.add_claim_def, "102--ADD--*--*--*"),
             (auth_map.edit_claim_def, "102--EDIT--*--*--*")]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_upgrade():
    rules = [(auth_map.start_upgrade, "109--ADD--action--*--start"),
             (auth_map.cancel_upgrade, "109--EDIT--action--start--cancel")]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_config():
    rules = [(auth_map.pool_config, "111--EDIT--action--*--*"), ]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_action():
    nym_rules = [(auth_map.pool_restart, "118--ADD--action--*--*"),
                 (auth_map.auth_rule, "120--EDIT--*--*--*"),
                 (auth_map.auth_rules, "122--EDIT--*--*--*"),
                 (auth_map.validator_info, "119--ADD--*--*--*")]

    for (rule, rule_str) in nym_rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_revoc_reg():
    nym_rules = [(auth_map.add_revoc_reg_def, "113--ADD--*--*--*"),
                 (auth_map.add_revoc_reg_entry, "114--ADD--*--*--*"),
                 (auth_map.edit_revoc_reg_def, "113--EDIT--*--*--*"),
                 (auth_map.edit_revoc_reg_entry, "114--EDIT--*--*--*")]

    for (rule, rule_str) in nym_rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()


def test_auth_map_disable_taa():
    rules = [(auth_map.disable_txn_author_agreement, '8--ADD--*--*--*')]

    for (rule, rule_str) in rules:
        assert rule.get_action_id() == rule_str
        assert rule_str in auth_map.auth_map.keys()



import pytest

from indy_common.constants import RS_CONTEXT_TYPE_VALUE, RS_ENCODING_TYPE_VALUE, RS_CRED_DEF_TYPE_VALUE, \
    RS_SCHEMA_TYPE_VALUE, RS_MAPPING_TYPE_VALUE, RS_PRES_DEF_TYPE_VALUE
from indy_common.types import Request
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.abstract_rich_schema_object_handler import \
    AbstractRichSchemaObjectHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.json_ld_context_handler import \
    JsonLdContextHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_cred_def_handler import \
    RichSchemaCredDefHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_encoding_handler import \
    RichSchemaEncodingHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_handler import RichSchemaHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_mapping_handler import \
    RichSchemaMappingHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_pres_def_handler import \
    RichSchemaPresDefHandler
from indy_node.test.helper import rich_schemas_enabled_scope
from indy_node.test.request_handlers.rich_schema.helper import context_request, rich_schema_request, \
    rich_schema_encoding_request, rich_schema_mapping_request, rich_schema_cred_def_request, \
    rich_schema_pres_def_request


@pytest.fixture(scope="module")
def tconf(tconf):
    with rich_schemas_enabled_scope(tconf):
        yield tconf


@pytest.fixture(params=[RS_CONTEXT_TYPE_VALUE, RS_SCHEMA_TYPE_VALUE,
                        RS_ENCODING_TYPE_VALUE, RS_MAPPING_TYPE_VALUE,
                        RS_CRED_DEF_TYPE_VALUE, RS_PRES_DEF_TYPE_VALUE])
def handler_and_request(request, db_manager, write_auth_req_validator) -> (AbstractRichSchemaObjectHandler, Request):
    if request.param == RS_CONTEXT_TYPE_VALUE:
        return JsonLdContextHandler(db_manager, write_auth_req_validator), context_request()
    if request.param == RS_SCHEMA_TYPE_VALUE:
        return RichSchemaHandler(db_manager, write_auth_req_validator), rich_schema_request()
    if request.param == RS_ENCODING_TYPE_VALUE:
        return RichSchemaEncodingHandler(db_manager, write_auth_req_validator), rich_schema_encoding_request()
    if request.param == RS_MAPPING_TYPE_VALUE:
        return RichSchemaMappingHandler(db_manager, write_auth_req_validator), rich_schema_mapping_request()
    if request.param == RS_CRED_DEF_TYPE_VALUE:
        return RichSchemaCredDefHandler(db_manager, write_auth_req_validator), rich_schema_cred_def_request()
    if request.param == RS_PRES_DEF_TYPE_VALUE:
        return RichSchemaPresDefHandler(db_manager, write_auth_req_validator), rich_schema_pres_def_request()


from indy_common.constants import CLAIM_DEF_SIGNATURE_TYPE, GET_CLAIM_DEF
from indy_common.req_utils import get_read_claim_def_from, get_read_claim_def_signature_type, \
    get_read_claim_def_schema_ref, get_read_claim_def_tag
from indy_node.server.request_handlers.domain_req_handlers.claim_def_handler import ClaimDefHandler

from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler


class GetClaimDefHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, GET_CLAIM_DEF, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        frm = get_read_claim_def_from(request)
        signature_type = get_read_claim_def_signature_type(request)
        schema_ref = get_read_claim_def_schema_ref(request)
        tag = get_read_claim_def_tag(request)
        keys, last_seq_no, last_update_time, proof = self.get_claim_def(
            author=frm,
            schema_seq_no=schema_ref,
            signature_type=signature_type,
            tag=tag
        )
        result = self.make_result(request=request,
                                  data=keys,
                                  last_seq_no=last_seq_no,
                                  update_time=last_update_time,
                                  proof=proof)
        result[CLAIM_DEF_SIGNATURE_TYPE] = signature_type
        return result

    def get_claim_def(self,
                      author: str,
                      schema_seq_no: str,
                      signature_type,
                      tag,
                      is_committed=True) -> (str, int, int, list):
        assert author is not None
        assert schema_seq_no is not None
        path = ClaimDefHandler.make_state_path_for_claim_def(author, schema_seq_no, signature_type, tag)
        try:
            keys, seq_no, last_update_time, proof = self.lookup(path, is_committed, with_proof=True)
            return keys, seq_no, last_update_time, proof
        except KeyError:
            return None, None, None, None



from .__metadata__ import (
    __title__, __version_info__, __version__, __manifest__, __description__,
    __long_description__, __keywords__, __url__, __author__,
    __author_email__, __maintainer__, __license__,
    load_version, set_version, load_manifest, set_manifest
)

PLUGIN_LEDGER_IDS = set()
PLUGIN_CLIENT_REQUEST_FIELDS = {}
PLUGIN_CLIENT_REQ_OP_TYPES = {}


# TODO review is it really necessary here
def setup_plugins():
    import sys
    import os
    import importlib_metadata
    import importlib    # noqa
    from importlib.util import module_from_spec, spec_from_file_location    # noqa: E402
    from indy_common.config_util import getConfigOnce   # noqa: E402
    import indy_node.server.plugin     # noqa: E402

    def find_and_load_plugin(plugin_name, plugin_root, installed_packages):
        if plugin_name in installed_packages:
            # TODO: Need a test for installed packages
            plugin_name = plugin_name.replace('-', '_')
            plugin = importlib.import_module(plugin_name)
        else:
            plugin_path = os.path.join(plugin_root.__path__[0],
                                       plugin_name, '__init__.py')
            spec = spec_from_file_location('__init__.py', plugin_path)
            plugin = module_from_spec(spec)
            spec.loader.exec_module(plugin)

        return plugin

    # TODO: Refactor to use plenum's setup_plugins
    # TODO: Should have a check to make sure no plugin defines any conflicting ledger id or request field
    global PLUGIN_LEDGER_IDS
    global PLUGIN_CLIENT_REQUEST_FIELDS
    global PLUGIN_CLIENT_REQ_OP_TYPES

    config = getConfigOnce(ignore_external_config_update_errors=True)

    plugin_root = config.PLUGIN_ROOT
    try:
        plugin_root = importlib.import_module(plugin_root)
    except ImportError:
        raise ImportError('Incorrect plugin root {}. No such package found'.
                          format(plugin_root))
    sys.path.insert(0, plugin_root.__path__[0])
    enabled_plugins = config.ENABLED_PLUGINS
    installed_packages = set(p.metadata["Name"] for p in importlib_metadata.distributions())
    for plugin_name in enabled_plugins:
        plugin = find_and_load_plugin(plugin_name, plugin_root, installed_packages)
        plugin_globals = plugin.__dict__
        # The following lines are idempotent so loading the same plugin twice is not the problem.
        if 'LEDGER_IDS' in plugin_globals:
            PLUGIN_LEDGER_IDS.update(plugin_globals['LEDGER_IDS'])
        if 'CLIENT_REQUEST_FIELDS' in plugin_globals:
            PLUGIN_CLIENT_REQUEST_FIELDS.update(plugin_globals['CLIENT_REQUEST_FIELDS'])
        if 'REQ_OP_TYPES' in plugin_globals:
            PLUGIN_CLIENT_REQ_OP_TYPES.update(plugin_globals['REQ_OP_TYPES'])

    # Reloading message types since some some schemas would have been changed
    import indy_common.types
    importlib.reload(indy_common.types)


try:
    import packaging
except ImportError:
    pass  # it is expected in raw env
else:
    setup_plugins()



MARKER_AUTH_RULE = "1"


def make_state_path_for_auth_rule(action_id) -> bytes:
    return "{MARKER}:{ACTION_ID}" \
        .format(MARKER=MARKER_AUTH_RULE,
                ACTION_ID=action_id).encode()



from typing import Optional

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.req_utils import get_txn_claim_def_schema_ref, get_txn_claim_def_tag, get_txn_claim_def_signature_type, \
    get_txn_claim_def_public_keys, get_write_claim_def_signature_type, get_write_claim_def_schema_ref, \
    get_write_claim_def_tag

from indy_common.constants import CLAIM_DEF, REF, SCHEMA, CLAIM_DEF_PUBLIC_KEYS, CLAIM_DEF_SCHEMA_REF
from indy_common.state.state_constants import MARKER_CLAIM_DEF

from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import get_request_data, get_from, get_seq_no, get_txn_time

from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler
from plenum.server.request_handlers.utils import encode_state_value


class ClaimDefHandler(WriteRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager, CLAIM_DEF, DOMAIN_LEDGER_ID)
        self.write_req_validator = write_req_validator

    def static_validation(self, request: Request):
        pass

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        # we can not add a Claim Def with existent ISSUER_DID
        # sine a Claim Def needs to be identified by seqNo
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        ref = operation[REF]
        try:
            txn = self.ledger.get_by_seq_no_uncommitted(ref)
        except KeyError:
            raise InvalidClientRequest(identifier,
                                       req_id,
                                       "Mentioned seqNo ({}) doesn't exist.".format(ref))
        if txn['txn']['type'] != SCHEMA:
            raise InvalidClientRequest(identifier,
                                       req_id,
                                       "Mentioned seqNo ({}) isn't seqNo of the schema.".format(ref))
        signature_type = get_write_claim_def_signature_type(request)
        schema_ref = get_write_claim_def_schema_ref(request)
        tag = get_write_claim_def_tag(request)

        path = self.make_state_path_for_claim_def(identifier, schema_ref, signature_type, tag)

        claim_def, _, _ = self.get_from_state(path, is_committed=False)

        if claim_def:
            self.write_req_validator.validate(request,
                                              [AuthActionEdit(txn_type=CLAIM_DEF,
                                                              field='*',
                                                              old_value='*',
                                                              new_value='*')])
        else:
            self.write_req_validator.validate(request,
                                              [AuthActionAdd(txn_type=CLAIM_DEF,
                                                             field='*',
                                                             value='*')])

    def gen_txn_id(self, txn):
        self._validate_txn_type(txn)
        path = self.prepare_claim_def_for_state(txn, path_only=True)
        return path.decode()

    def update_state(self, txn, prev_result, request, is_committed=False) -> None:
        self._validate_txn_type(txn)
        path, value_bytes = self.prepare_claim_def_for_state(txn)
        self.state.set(path, value_bytes)
        return txn

    @staticmethod
    def prepare_claim_def_for_state(txn, path_only=False):
        origin = get_from(txn)
        schema_seq_no = get_txn_claim_def_schema_ref(txn)
        if schema_seq_no is None:
            raise ValueError("'{}' field is absent, "
                             "but it must contain schema seq no".format(CLAIM_DEF_SCHEMA_REF))
        data = get_txn_claim_def_public_keys(txn)
        if data is None:
            raise ValueError("'{}' field is absent, "
                             "but it must contain components of keys"
                             .format(CLAIM_DEF_PUBLIC_KEYS))
        signature_type = get_txn_claim_def_signature_type(txn)
        tag = get_txn_claim_def_tag(txn)
        path = ClaimDefHandler.make_state_path_for_claim_def(origin, schema_seq_no, signature_type, tag)
        if path_only:
            return path
        seq_no = get_seq_no(txn)
        txn_time = get_txn_time(txn)
        value_bytes = encode_state_value(data, seq_no, txn_time)
        return path, value_bytes

    @staticmethod
    def make_state_path_for_claim_def(authors_did, schema_seq_no, signature_type, tag) -> bytes:
        return "{DID}:{MARKER}:{SIGNATURE_TYPE}:{SCHEMA_SEQ_NO}:{TAG}" \
            .format(DID=authors_did,
                    MARKER=MARKER_CLAIM_DEF,
                    SIGNATURE_TYPE=signature_type,
                    SCHEMA_SEQ_NO=schema_seq_no,
                    TAG=tag).encode()



import json

import pytest

from indy_common.constants import CRED_DEF_ID, CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_SIGNATURE_TYPE, \
    CLAIM_DEF_TAG, VALUE, ISSUANCE_TYPE, REVOKED, PREV_ACCUM, ISSUANCE_BY_DEFAULT
from indy_common.state.domain import make_state_path_for_claim_def
from indy_node.test.anon_creds.conftest import build_revoc_reg_entry_for_given_revoc_reg_def
from plenum.common.exceptions import RequestNackedException
from plenum.test.helper import sdk_sign_request_from_dict, sdk_send_and_check


def test_incorrect_revoc_reg_def(looper,
                                  txnPoolNodeSet,
                                  sdk_wallet_steward,
                                  sdk_pool_handle,
                                  send_claim_def,
                                  build_revoc_def_by_default):
    _, author_did = sdk_wallet_steward
    claim_def_req = send_claim_def[0]
    revoc_reg = build_revoc_def_by_default
    revoc_reg['operation'][CRED_DEF_ID] = \
        make_state_path_for_claim_def(author_did,
                                      str(claim_def_req['operation'][CLAIM_DEF_SCHEMA_REF]),
                                      claim_def_req['operation'][CLAIM_DEF_SIGNATURE_TYPE],
                                      claim_def_req['operation'][CLAIM_DEF_TAG]
                                      ).decode()

    # test incorrect ISSUANCE_TYPE
    revoc_reg['operation'][VALUE][ISSUANCE_TYPE] = "incorrect_type"
    revoc_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, revoc_reg['operation'])
    with pytest.raises(RequestNackedException, match='unknown value'):
        sdk_send_and_check([json.dumps(revoc_req)], looper, txnPoolNodeSet, sdk_pool_handle)

    # test correct ISSUANCE_TYPE
    revoc_reg['operation'][VALUE][ISSUANCE_TYPE] = ISSUANCE_BY_DEFAULT
    revoc_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, revoc_reg['operation'])
    sdk_send_and_check([json.dumps(revoc_req)], looper, txnPoolNodeSet, sdk_pool_handle)

    # send revoc_reg_entry to check that revoc_reg_def ordered correctly
    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_req)
    rev_reg_entry[VALUE][REVOKED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, rev_reg_entry)
    sdk_send_and_check([json.dumps(rev_entry_req)], looper, txnPoolNodeSet, sdk_pool_handle)



import functools
import json
import multiprocessing
import os
import socket
import subprocess
from datetime import datetime
from typing import List, Tuple

import base58
import dateutil.tz
from plenum.bls.bls_crypto_factory import create_default_bls_crypto_factory
from plenum.test.node_catchup.helper import waitNodeDataEquality, ensure_all_nodes_have_same_data

from plenum.common.keygen_utils import init_bls_keys

from indy.ledger import build_pool_upgrade_request
from plenum.common.constants import DATA, VERSION, FORCE
from plenum.common.txn_util import get_type, get_payload_data, get_from
from plenum.common.util import randomString, hexToFriendly
from plenum.test import waits as plenumWaits
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request, sdk_send_update_node, \
    sdk_pool_refresh, sdk_add_new_nym
from stp_core.common.log import getlogger
from stp_core.loop.eventually import eventually

from indy_common.constants import NODE_UPGRADE, ACTION, \
    UPGRADE_MESSAGE, MESSAGE_TYPE, APP_NAME
from indy_common.config import controlServiceHost, controlServicePort
import indy_node
from indy_node.server.upgrade_log import UpgradeLogData, UpgradeLog
from indy_node.server.upgrader import Upgrader
from indy_node.test.helper import TestNode, TestNodeBootstrap
from indy_node.utils.node_control_tool import NodeControlTool
from indy_common.config_helper import NodeConfigHelper

logger = getlogger()


class TestNodeNoProtocolVersion(TestNode):
    def processNodeRequest(self, request, frm):
        if request.protocolVersion is not None:
            raise ValueError('Do not understand what protocolVersion is!!!')
        super().processNodeRequest(request, frm)


def sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_trustee, upgrade_data):
    _, did = sdk_wallet_trustee
    req = get_req_from_update(looper, did, upgrade_data)
    return sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                              sdk_pool_handle, req)


def sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgrade_data):
    req = sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_trustee, upgrade_data)
    sdk_get_and_check_replies(looper, [req])


def get_req_from_update(looper, did, nup):
    req = looper.loop.run_until_complete(
        build_pool_upgrade_request(did, nup['name'], nup['version'], nup['action'], nup['sha256'],
                                   nup['timeout'],
                                   json.dumps(nup['schedule']) if 'schedule' in nup else None,
                                   nup['justification'] if 'justification' in nup else 'null',
                                   nup['reinstall'] if 'reinstall' in nup else None,
                                   nup[FORCE] if FORCE in nup else None,
                                   nup.get('package', None)))
    return req


def clear_aq_stash(nodes):
    for node in nodes:
        node.upgrader.aqStash.clear()


def checkUpgradeScheduled(nodes, version: str, schedule=None):
    for node in nodes:
        assert len(node.upgrader.aqStash) == 1
        assert node.upgrader.scheduledAction
        assert node.upgrader.scheduledAction.version.full == version
        if schedule:
            assert node.upgrader.scheduledAction.when == \
                   dateutil.parser.parse(schedule[node.id])


def checkNoUpgradeScheduled(nodes):
    for node in nodes:
        assert len(node.upgrader.aqStash) == 0
        assert node.upgrader.scheduledAction is None


def codeVersionInfo():
    return indy_node.__version_info__.parts


def releaseVersion():
    return '.'.join(map(str, codeVersionInfo()[:3]))


def bumpVersion(v):
    parts = v.split('.')
    return '.'.join(parts[:-1] + [str(int(parts[-1]) + 1)])


def lowerVersion(v):
    parts = v.split('.')
    for i in reversed(range(len(parts))):
        if int(parts[i]) > 0:
            parts[i] = str(int(parts[i]) - 1)
            break
    else:
        raise ValueError('Version {} cannot be lowered'.format(v))
    return '.'.join(parts)


def bumpedVersion(ver=None):
    return bumpVersion(ver or releaseVersion())


def loweredVersion():
    return lowerVersion(releaseVersion())


class NodeControlToolPatched(NodeControlTool):

    def __init__(self, patch_f, *args, **kwargs):
        patch_f(self)
        super().__init__(*args, **kwargs)


class NodeControlToolExecutor:
    def __init__(self, backup_dir, backup_target, transform=lambda tool: None, config=None):
        self.tool = NodeControlTool(backup_dir=backup_dir, backup_target=backup_target, config=config)
        transform(self.tool)
        self.p = multiprocessing.Process(target=self.tool.start)
        self.p.start()
        logger.debug("NCTProcess was started with pid: {}".format(self.p.pid))

    def stop(self):
        logger.debug("Send stop to NCTProcess with pid: {}".format(self.p.pid))
        self.tool.server.close()
        self.p.terminate()
        # check that process with NodeControlTool.start function really stop.
        # process.terminate() just send SIGTERM and is not guarantee that process stops
        if self.p.is_alive():
            logger.debug("NCTProcess still alive, with pid: {}".format(self.p.pid))
            # while process is still alive, join with main process and wait

            # FIXME: here was self.p.join(3), but since we've added (ok, Andrew Nikitin
            # has added) handler for SIGTERM here we wait for child process infinitely,
            # but for now we have no time to fix it in more elegant way as it is not a
            # real situation (Andrew said), so he has proposed this ugly hack.
            os.kill(self.p.pid, 9)
        logger.debug("NCTProcess must be stopped, with pid: {}".format(self.p.pid))


def composeUpgradeMessage(version, pkg_name: str = APP_NAME):
    return (json.dumps({"version": version, "pkg_name": pkg_name, MESSAGE_TYPE: UPGRADE_MESSAGE})).encode()


def sendUpgradeMessage(version, pkg_name: str = APP_NAME):
    sock = socket.create_connection(
        (controlServiceHost, controlServicePort))
    sock.sendall(composeUpgradeMessage(version, pkg_name=pkg_name))
    sock.close()


def nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout):
    ret = type("", (), {})()
    ret.returncode = 0
    ret.stdout = stdout if isinstance(stdout, bytes) else stdout.encode()
    tool.base_dir = tdir
    tool.indy_dir = os.path.join(tool.base_dir, '.indy')
    tool.tmp_dir = os.path.join(tool.base_dir, '.indy_tmp')
    if not os.path.exists(tool.indy_dir):
        os.mkdir(tool.indy_dir)
    if not os.path.exists(tool.tmp_dir):
        os.mkdir(tool.tmp_dir)
    monkeypatch.setattr(subprocess, 'run', lambda *x, **y: ret)
    monkeypatch.setattr(tool, '_do_migration', lambda *x: None)


def get_valid_code_hash():
    return randomString(64)


def populate_log_with_upgrade_events(
        pool_txn_node_names, tdir, tconf, version: Tuple[str, str, str], pkg_name: str = APP_NAME):
    for nm in pool_txn_node_names:
        config_helper = NodeConfigHelper(nm, tconf, chroot=tdir)
        ledger_dir = config_helper.ledger_dir
        os.makedirs(ledger_dir)
        log = UpgradeLog(os.path.join(ledger_dir, tconf.upgradeLogFile))
        when = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
        ev_data = UpgradeLogData(when, version, randomString(10), pkg_name)
        log.append_scheduled(ev_data)
        log.append_started(ev_data)


def check_node_sent_acknowledges_upgrade(
        looper, node_set, node_ids, allowed_actions: List, ledger_size, expected_version):
    '''
    Check that each node has sent NODE_UPGRADE txn with the specified actions
    '''
    check = functools.partial(
        check_ledger_after_upgrade,
        node_set,
        allowed_actions,
        ledger_size,
        expected_version,
        node_ids=node_ids)

    timeout = plenumWaits.expectedTransactionExecutionTime(len(node_set))
    looper.run(
        eventually(
            check,
            retryWait=1,
            timeout=timeout))


def emulate_restart_pool_for_upgrade(nodes):
    for node in nodes:
        bootstrap = TestNodeBootstrap(node)
        node.upgrader = bootstrap.init_upgrader()
        node.acknowledge_upgrade()


def emulate_view_change_pool_for_upgrade(nodes):
    for node in nodes:
        node.upgrader.processLedger()
        node.acknowledge_upgrade()


def check_node_do_not_sent_acknowledges_upgrade(
        looper, node_set, node_ids, allowed_actions: List, ledger_size, expected_version):
    '''
    Check that each node has sent NODE_UPGRADE txn with the specified actions
    '''
    looper.runFor(5)
    check_ledger_after_upgrade(node_set, allowed_actions,
                               ledger_size, expected_version,
                               node_ids=node_ids)


def clear_config_ledger(node_set):
    for node in node_set:
        node.configLedger.reset_uncommitted()
        node.configLedger._transactionLog.reset()
        node.configLedger.tree.reset()


def check_ledger_after_upgrade(
        node_set,
        allowed_actions,
        ledger_size,
        expected_version,
        allowed_txn_types=[NODE_UPGRADE],
        node_ids=None):
    versions = set()
    for node in node_set:
        print(len(node.configLedger))
        assert len(node.configLedger) == ledger_size
        ids = set()
        for _, txn in node.configLedger.getAllTxn():
            typ = get_type(txn)
            assert typ in allowed_txn_types
            txn_data = get_payload_data(txn)
            data = txn_data
            if typ == NODE_UPGRADE:
                data = txn_data[DATA]

            assert data[ACTION]
            assert data[ACTION] in allowed_actions
            ids.add(get_from(txn))

            assert data[VERSION]
            versions.add(data[VERSION])
        ids.add(node.id)

        if node_ids:
            assert ids == set(node_ids)
    assert len(versions) == 1
    assert list(versions)[0] == expected_version


def check_no_loop(nodeSet, ev_type, pkg_name: str = APP_NAME):
    for node in nodeSet:
        # mimicking upgrade start
        node.upgrader._actionLog.append_started(
            UpgradeLogData(
                datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc()),
                node.upgrader.scheduledAction.version,
                node.upgrader.scheduledAction.upgrade_id,
                pkg_name
            )
        )
        node.notify_upgrade_start()
        # mimicking upgrader's initialization after restart
        node.upgrader.process_action_log_for_first_run()

        node.upgrader.scheduledAction = None
        assert node.upgrader._actionLog.last_event.ev_type == ev_type
        # mimicking node's catchup after restart
        node.postConfigLedgerCaughtUp()
        assert node.upgrader.scheduledAction is None
        assert node.upgrader._actionLog.last_event.ev_type == ev_type


def sdk_change_bls_key(looper, txnPoolNodeSet,
                       node,
                       sdk_pool_handle,
                       sdk_wallet_steward,
                       add_wrong=False,
                       new_bls=None,
                       new_key_proof=None):
    if add_wrong:
        _, new_blspk, key_proof = create_default_bls_crypto_factory().generate_bls_keys()
    else:
        new_blspk, key_proof = init_bls_keys(node.keys_dir, node.name)
    key_in_txn = new_bls or new_blspk
    bls_key_proof = new_key_proof or key_proof
    node_dest = hexToFriendly(node.nodestack.verhex)
    sdk_send_update_node(looper, sdk_wallet_steward,
                         sdk_pool_handle,
                         node_dest, node.name,
                         None, None,
                         None, None,
                         bls_key=key_in_txn,
                         services=None,
                         key_proof=bls_key_proof)
    poolSetExceptOne = list(txnPoolNodeSet)
    poolSetExceptOne.remove(node)
    waitNodeDataEquality(looper, node, *poolSetExceptOne)
    sdk_pool_refresh(looper, sdk_pool_handle)
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_steward,
                    alias=randomString(5))
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet)
    return new_blspk


def count_action_log_entries(upg_log, func):
    ret = 0
    for upg_rec in upg_log:
        if func(upg_rec):
            ret += 1
    return ret


def count_action_log_package(upg_log, pkg_name):
    return count_action_log_entries(upg_log, lambda entry: entry.data.pkg_name == pkg_name)



from copy import deepcopy

import pytest

from plenum.common.exceptions import RequestRejectedException
from plenum.common.constants import VERSION

from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent, lowerVersion


def test_do_not_upgrade_to_the_same_version(looper, tconf, nodeSet,
                                            validUpgrade, sdk_pool_handle,
                                            sdk_wallet_trustee):
    upgr1 = deepcopy(validUpgrade)
    upgr1[VERSION] = lowerVersion(validUpgrade['version'])

    # An upgrade is not scheduled
    with pytest.raises(RequestRejectedException) as ex:
        sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr1)
    ex.match("Version {} is not upgradable".format(upgr1[VERSION]))



from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_node.test.auth_rule.helper import create_verkey_did, sdk_send_and_check_auth_rule_request
from plenum.common.constants import STEWARD, NYM, ROLE, STEWARD_STRING
from plenum.test import waits
from plenum.test.delayers import cDelay
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.stasher import delay_rules


def test_use_modified_rules_from_uncommitted(looper,
                                             txnPoolNodeSet,
                                             sdk_wallet_trustee,
                                             sdk_wallet_steward,
                                             sdk_pool_handle):
    node_stashers = [n.nodeIbStasher for n in txnPoolNodeSet]
    wh, _ = sdk_wallet_trustee
    new_steward_did, new_steward_verkey = create_verkey_did(looper, wh)
    changed_constraint = AuthConstraint(role=STEWARD,
                                        sig_count=1)
    with delay_rules(node_stashers, cDelay()):
        r_auth = sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                                      auth_action=ADD_PREFIX,
                                                      auth_type=NYM, field=ROLE, new_value=STEWARD, old_value=None,
                                                      constraint=changed_constraint.as_dict, no_wait=True)
        looper.runFor(waits.expectedPrePrepareTime(len(txnPoolNodeSet)))
        r_add_steward = sdk_add_new_nym(looper,
                                        sdk_pool_handle,
                                        sdk_wallet_steward,
                                        'newSteward2',
                                        STEWARD_STRING,
                                        dest=new_steward_did, verkey=new_steward_verkey,
                                        no_wait=True)

    sdk_get_and_check_replies(looper, [r_auth])
    sdk_get_and_check_replies(looper, [r_add_steward])



import json

import pytest
from indy.ledger import build_get_schema_request, parse_get_schema_response

from indy_common.state.domain import make_state_path_for_claim_def
from plenum.common.types import OPERATION
from plenum.common.util import randomString

from indy_common.authorize.auth_actions import ADD_PREFIX, AuthActionAdd, AuthActionEdit, EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import CLAIM_DEF, ID, REVOC_TYPE, TAG, CRED_DEF_ID, VALUE, ISSUANCE_TYPE, MAX_CRED_NUM, \
    TAILS_HASH, TAILS_LOCATION, PUBLIC_KEYS, ISSUANCE_BY_DEFAULT, REVOC_REG_DEF, CLAIM_DEF_SCHEMA_REF, \
    CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_TAG
from indy_node.test.anon_creds.conftest import send_revoc_reg_def_by_default
from indy_node.test.api.helper import sdk_write_schema
from indy_node.test.auth_rule.auth_framework.basic import AbstractTest, AuthTest
from indy_node.test.claim_def.test_send_claim_def import sdk_send_claim_def
from plenum.common.constants import TRUSTEE, TXN_TYPE
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import sdk_get_and_check_replies, \
    sdk_multi_sign_request_objects, sdk_send_signed_requests, sdk_sign_and_submit_req, sdk_get_reply, \
    sdk_sign_request_from_dict, sdk_send_and_check, sdk_sign_and_submit_op
from plenum.test.pool_transactions.helper import sdk_add_new_nym, sdk_sign_and_send_prepared_request
from indy_common.authorize import auth_map

from indy_node.test.helper import build_auth_rule_request_json


def get_schema_json(looper, sdk_pool_handle, sdk_wallet_trustee):
    wallet_handle, identifier = sdk_wallet_trustee
    schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_trustee)
    schema_id = json.loads(schema_json)['id']

    request = looper.loop.run_until_complete(build_get_schema_request(identifier, schema_id))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request))[1]
    _, schema_json = looper.loop.run_until_complete(parse_get_schema_response(json.dumps(reply)))
    return schema_json


class AddRevocRegDefTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.claim_def_req = None

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        self.claim_def_req = self.get_claim_def()[0][0]

    def run(self):

        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            self.send_revoc_reg_def(self.claim_def_req, wallet=self.trustee_wallet)

        # Step 3. Check, that new auth rule is used
        self.send_revoc_reg_def(self.claim_def_req, wallet=self.new_default_wallet)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.send_revoc_reg_def(self.claim_def_req, wallet=self.trustee_wallet)

    def result(self):
        pass

    def get_claim_def(self):
        schema_json = get_schema_json(self.looper, self.sdk_pool_handle, self.trustee_wallet)
        return sdk_send_claim_def(self.looper, self.sdk_pool_handle, self.trustee_wallet, 'tag' + randomString(4),
                                  schema_json)

    def send_revoc_reg_def(self, claim_def_req=None, author_did=None, wallet=None, revoc_reg_def=None):
        claim_def_req = claim_def_req if claim_def_req else self.claim_def_req
        wallet = wallet or self.trustee_wallet
        req = revoc_reg_def if revoc_reg_def else self.build_revoc_def_by_default(claim_def_req,
                                                                                  author_did=author_did,
                                                                                  wallet=wallet)
        return sdk_send_and_check([json.dumps(req)], self.looper, self.env.txnPoolNodeSet, self.sdk_pool_handle)

    def build_revoc_def_by_default(self, claim_def_req, author_did=None, wallet=None, req_def_id=None):
        wallet = wallet if wallet else self.trustee_wallet
        author_did = author_did if author_did else self.trustee_wallet[1]
        data = {
            ID: req_def_id if req_def_id else randomString(50),
            TXN_TYPE: REVOC_REG_DEF,
            REVOC_TYPE: "CL_ACCUM",
            TAG: randomString(5),
            CRED_DEF_ID: make_state_path_for_claim_def(author_did,
                                                       str(claim_def_req['operation'][CLAIM_DEF_SCHEMA_REF]),
                                                       claim_def_req['operation'][CLAIM_DEF_SIGNATURE_TYPE],
                                                       claim_def_req['operation'][CLAIM_DEF_TAG]
                                                       ).decode(),
            VALUE: {
                ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
                MAX_CRED_NUM: 1000000,
                TAILS_HASH: randomString(50),
                TAILS_LOCATION: 'http://tails.location.com',
                PUBLIC_KEYS: {},
            }
        }
        req = sdk_sign_request_from_dict(self.looper, wallet, data)
        return req

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=REVOC_REG_DEF,
            field='*',
            new_value='*',
            constraint=constraint.as_dict
        )


class EditRevocRegDefTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.revoc_reg_def_trustee = None
        self.revoc_reg_def_new = None
        self.add_revoc_reg_def = AddRevocRegDefTest(env, auth_map.add_revoc_reg_def.get_action_id())
        self.add_revoc_reg_def.prepare()

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        self.revoc_reg_def_trustee = self.add_revoc_reg_def.send_revoc_reg_def()[0][0]
        self.add_revoc_reg_def.send_and_check(self.add_revoc_reg_def.changed_auth_rule, self.trustee_wallet)
        self.revoc_reg_def_new = self.add_revoc_reg_def.send_revoc_reg_def(wallet=self.new_default_wallet)[0][0]

    def do_edit_revoc_reg_def(self, origin_rrd_req, wallet):
        origin_rrd_req[OPERATION][VALUE][TAILS_HASH] = randomString(20)
        resp = sdk_sign_and_submit_op(self.looper, self.sdk_pool_handle, wallet, op=origin_rrd_req[OPERATION])
        sdk_get_and_check_replies(self.looper, [resp])

    def run(self):
        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            self.do_edit_revoc_reg_def(self.revoc_reg_def_trustee, wallet=self.trustee_wallet)

        # Step 3. Check, that new auth rule is used
        self.do_edit_revoc_reg_def(self.revoc_reg_def_new, self.new_default_wallet)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.do_edit_revoc_reg_def(self.revoc_reg_def_trustee, self.trustee_wallet)

        self.teardown_for_add()

    def teardown_for_add(self):
        self.add_revoc_reg_def.send_and_check(self.add_revoc_reg_def.default_auth_rule, self.trustee_wallet)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = self.add_revoc_reg_def.new_default_wallet
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=EDIT_PREFIX,
            auth_type=REVOC_REG_DEF,
            field='*',
            old_value='*',
            new_value='*',
            constraint=constraint.as_dict
        )



from indy_common.constants import POOL_CONFIG, WRITES
from plenum.common.txn_util import get_type, get_payload_data

from stp_core.common.log import getlogger

logger = getlogger()


class PoolConfig:
    def __init__(self, ledger):
        self.writes = True
        self.ledger = ledger

    def isWritable(self):
        return self.writes

    def handleConfigTxn(self, txn) -> None:
        """
        Handles transaction of type POOL_CONFIG

        :param txn:
        """
        if get_type(txn) == POOL_CONFIG:
            self.writes = get_payload_data(txn)[WRITES]

    # TODO: config ledger is read from the start to the end. Think about optimization
    # TODO: PoolConfig and Updater both read config ledger independently
    def processLedger(self) -> None:
        """
        Checks ledger config txns and perfomes recent one

        :return:
        """
        logger.debug('{} processing config ledger for any POOL_CONFIGs'.format(
            self), extra={"tags": ["pool-config"]})
        for _, txn in self.ledger.getAllTxn():
            if get_type(txn) == POOL_CONFIG:
                self.handleConfigTxn(txn)



import json
from random import randint
from typing import Any, Dict, Optional, Union

from indy_common.constants import GET_NYM, NYM


def build_nym_request(
    identifier: str,
    dest: str,
    verkey: Optional[str] = None,
    diddoc_content: Optional[Union[dict, str]] = None,
    role: Optional[str] = None,
    version: Optional[int] = None,
):
    request = {
        "identifier": identifier,
        "reqId": randint(100, 1000000),
        "protocolVersion": 2
    }

    operation: Dict[str, Any] = {
        "dest": dest,
        "type": NYM
    }

    if verkey:
        operation["verkey"] = verkey
    if diddoc_content:
        operation["diddocContent"] = (
            json.dumps(diddoc_content)
            if isinstance(diddoc_content, dict)
            else diddoc_content
        )
    if role:
        operation["role"] = role

    if version:
        operation["version"] = version

    request["operation"] = operation
    return json.dumps(request)


def build_get_nym_request(
    identifier: str,
    dest: str,
    timestamp: Optional[int] = None,
    seq_no: Optional[int] = None,
):
    request = {
        "identifier": identifier,
        "reqId": randint(100, 1000000),
        "protocolVersion": 2
    }

    operation: Dict[str, Any] = {
        "dest": dest,
        "type": GET_NYM
    }

    if timestamp:
        operation["timestamp"] = timestamp

    if seq_no:
        operation["seqNo"] = seq_no

    request["operation"] = operation
    return json.dumps(request)



import base58
from plenum.common.constants import DOMAIN_LEDGER_ID

from plenum.server.batch_handlers.batch_request_handler import BatchRequestHandler
from plenum.server.batch_handlers.three_pc_batch import ThreePcBatch
from plenum.server.database_manager import DatabaseManager


class IdrCacheBatchHandler(BatchRequestHandler):
    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, DOMAIN_LEDGER_ID)

    def commit_batch(self, three_pc_batch: ThreePcBatch, prev_handler_result=None):
        state_root = base58.b58decode(three_pc_batch.state_root.encode())
        self.database_manager.idr_cache.onBatchCommitted(state_root)

    def post_batch_applied(self, three_pc_batch: ThreePcBatch, prev_handler_result=None):
        self.database_manager.idr_cache.currentBatchCreated(three_pc_batch.state_root,
                                                            three_pc_batch.pp_time)

    def post_batch_rejected(self, ledger_id, prev_handler_result=None):
        self.database_manager.idr_cache.batchRejected()



import asyncio

import multiprocessing

import time

import pytest

from indy_common.constants import POOL_RESTART, RESTART_MESSAGE
from indy_node.test.pool_restart.helper import compose_restart_message, \
    send_restart_message
from stp_core.loop.eventually import eventually
from indy_node.test.upgrade.helper import NodeControlToolExecutor as NCT, \
    nodeControlGeneralMonkeypatching

m = multiprocessing.Manager()
whitelist = ['Unexpected error in _restart test']


@pytest.fixture(scope='module')
def tconf(tconf, tdir):
    tconf.LOG_DIR = tdir
    yield tconf


def test_node_control_tool_restart(looper, tdir, monkeypatch, tconf):
    received = m.list()
    msg = RESTART_MESSAGE
    stdout = 'teststdout'

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        monkeypatch.setattr(tool, '_process_data', received.append)

    def check_message():
        assert len(received) == 1
        assert received[0] == compose_restart_message(msg)

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform, config=tconf)
    try:
        send_restart_message(msg)
        looper.run(eventually(check_message))
    finally:
        nct.stop()


def test_communication_with_node_control_tool(looper, tdir, tconf, monkeypatch):
    received = m.list()
    msg = RESTART_MESSAGE
    stdout = 'teststdout'

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        monkeypatch.setattr(tool, '_restart', restart_count)

    def check_restart_count():
        assert len(received) == 1

    def restart_count():
        received.append(RESTART_MESSAGE)

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform, config=tconf)
    try:
        send_restart_message(msg)
        looper.run(eventually(check_restart_count))
    finally:
        nct.stop()



from typing import Iterable

try:
    from common.version import (
        InvalidVersionError, PEP440BasedVersion, SemVerBase,
        SourceVersion, PackageVersion
    )
except ImportError:
    from indy_common.node_version_fallback import (
        InvalidVersionError, NodeVersionFallback
    )
    NodeVersion = NodeVersionFallback
else:
    # TODO inherit PlenumVersion once plenum is merged to node
    class NodeVersion(
        PEP440BasedVersion, SemVerBase, SourceVersion, PackageVersion
    ):
        def __init__(self, version: str, **kwargs):
            super().__init__(version, **kwargs)

            # additional restrictions
            if self._version.pre:
                if self._version.pre[0] != 'rc':
                    raise InvalidVersionError(
                        "pre-release phase '{}' is unexpected"
                        .format(self._version.pre[0])
                    )

            if self._version.post:
                raise InvalidVersionError("post-release is unexpected")

            if self._version.epoch:
                raise InvalidVersionError("epoch is unexpected")

            if self._version.local:
                raise InvalidVersionError("local version part is unexpected")

            # TODO set constraint to 3 only once new release logic becomes completed
            if len(self.release_parts) not in (2, 3):
                # TODO and here
                raise InvalidVersionError(
                    "release part should contain only 2 or 3 parts")

        @property
        def parts(self) -> Iterable:
            return super().parts[1:6]

        @property
        def upstream(self) -> SourceVersion:
            """ Upstream part of the package. """
            return self
finally:
    InvalidVersionError = InvalidVersionError



import json
from base64 import b64encode
from binascii import hexlify
from hashlib import sha256

import pytest
from indy.did import create_and_store_my_did
from libnacl import randombytes
from libnacl.secret import SecretBox

from indy_node.test.helper import sdk_add_attribute_and_check
from plenum.common.exceptions import RequestRejectedException, RequestNackedException
from plenum.common.util import rawToFriendly, randomString
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def testSendAttribSucceedsForExistingDest(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({'name': 'Alice'})
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribFailsForNotExistingDest(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    wh, _ = sdk_wallet_trustee
    seed = randomString(32)
    did, _ = looper.loop.run_until_complete(create_and_store_my_did(
        wh, json.dumps({'seed': seed})))

    parameters = json.dumps({'name': 'Alice'})
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_trustee,
                                    parameters, dest=did)
    e.match('dest should be added before adding attribute for it')


def testSendAttribSucceedsForRawWithCompoundAttr(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'dateOfBirth': {
            'year': 1984,
            'month': 5,
            'dayOfMonth': 23
        }
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithNullifiedAttr(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'name': None
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithEndpointWithHaContainingIpAddrAndPort(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:6321'
        }
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithEndpointWithHaBeingNull(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': None
        }
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithEndpointWithValidHaAndOtherProperties(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:6321',
            'name': 'SOV Agent',
            'description': 'The SOV agent.'
        }
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithEndpointWithoutHaButWithOtherProperties(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'name': 'SOV Agent',
            'description': 'The SOV agent.'
        }
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithEndpointWithoutProperties(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {}
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribSucceedsForRawWithEndpointBeingNull(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': None
    })
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)


def testSendAttribFailsForRawWithEndpointWithHaIfIpAddrHasWrongFormat(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117:6321'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint address')


def testSendAttribFailsForRawWithEndpointWithHaIfSomeIpComponentsAreNegative(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.-1.117.186:6321'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint address')


def testSendAttribFailsForRawWithEndpointWithHaIfSomeIpCompHigherThanUpperBound(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.256.186:6321'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint address')


def testSendAttribFailsForRawWithEndpointWithHaIfIpAddrIsEmpty(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': ':6321'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint address')


def testSendAttribFailsForRawWithEndpointWithHaIfPortIsNegative(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:-1'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint port')


def testSendAttribFailsForRawWithEndpointWithHaIfPortIsHigherThanUpperBound(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:65536'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint port')


def testSendAttribFailsForRawWithEndpointWithHaIfPortIsFloat(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:6321.5'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint port')


def testSendAttribFailsForRawWithEndpointWithHaIfPortHasWrongFormat(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:ninety'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint port')


def testSendAttribFailsForRawWithEndpointWithHaIfPortIsEmpty(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186:'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint port')


def testSendAttribFailsForRawWithEndpointWithHaContainingIpAddrOnly(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '52.11.117.186'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint format')


def testSendAttribFailsForRawWithEndpointWithHaContainingPortOnly(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': '6321'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint format')


def testSendAttribFailsForRawWithEndpointWithHaContainingDomainNameAndPort(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': 'sovrin.org:6321'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint address')


def testSendAttribFailsForRawWithEndpointWithHaContainingDomainNameOnly(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': 'sovrin.org'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint format')


def testSendAttribFailsForRawWithEndpointWithHaBeingHumanReadableText(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': 'This is not a host address.'
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint format')


def testSendAttribFailsForRawWithEndpointWithHaBeingDecimalNumber(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': 42
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('is not iterable')


def testSendAttribFailsForRawWithEndpointWithEmptyHa(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': {
            'ha': ''
        }
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('invalid endpoint format')


def testSendAttribFailsForRawWithEndpointBeingEmptyString(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'endpoint': ''
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match('should be a dict')


def testSendAttribFailsIfRawContainsMulipleAttrs(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({
        'name': 'Alice',
        'dateOfBirth': '05/23/2017'
    })
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match(' should contain one attribute')


def testSendAttribFailsIfRawContainsNoAttrs(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = json.dumps({})
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters)
    e.match(' should contain one attribute')


def testSendAttribSucceedsForHexSha256Hash(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })

    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                xhash=sha256(raw.encode()).hexdigest())


def testSendAttribSucceedsForHexHashWithLettersInBothCases(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                xhash='6d4a333838d0ef96756cccC680AF2531075C512502Fb68c5503c63d93de859b3')


def testSendAttribFailsForHashShorterThanSha256(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=hexlify(randombytes(31)).decode())
    e.match('not a valid hash')


def testSendAttribFailsForHashLongerThanSha256(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=hexlify(randombytes(33)).decode())
    e.match('not a valid hash')


def testSendAttribFailsForBase58Hash(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })
    hash = sha256(raw.encode()).digest()

    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=rawToFriendly(hash))
    e.match('not a valid hash')


def testSendAttribFailsForBase64Hash(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })

    hash = sha256(raw.encode()).digest()
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=b64encode(hash).decode())
    e.match('not a valid hash')


def testSendAttribHasInvalidSyntaxIfHashIsEmpty(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash='')
    e.match('not a valid hash')


def testSendAttribSucceedsForNonEmptyEnc(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })
    secretBox = SecretBox()

    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                enc=secretBox.encrypt(raw.encode()).hex())


def testSendAttribHasInvalidSyntaxIfEncIsEmpty(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    enc='')
    e.match('empty string')


def testSendAttribHasInvalidSyntaxIfRawAndHashPassedAtSameTime(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })

    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=sha256(raw.encode()).hexdigest(), enc=raw)
    e.match('only one field from raw, enc, hash is expected')


def testSendAttribHasInvalidSyntaxIfRawAndEncPassedAtSameTime(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })
    secretBox = SecretBox()

    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=secretBox.encrypt(raw.encode()).hex(), enc=raw)
    e.match('not a valid hash')


def testSendAttribHasInvalidSyntaxIfHashAndEncPassedAtSameTime(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })

    secretBox = SecretBox()
    encryptedRaw = secretBox.encrypt(raw.encode())

    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = None
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=sha256(encryptedRaw).hexdigest(), enc=encryptedRaw.hex())
    e.match('only one field from raw, enc, hash is expected')


def testSendAttribHasInvalidSyntaxIfRawHashAndEncPassedAtSameTime(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    raw = json.dumps({
        'name': 'Alice'
    })

    secretBox = SecretBox()
    encryptedRaw = secretBox.encrypt(raw.encode())
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    parameters = raw
    with pytest.raises(RequestNackedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, new_wallet, parameters,
                                    xhash=sha256(encryptedRaw).hexdigest(), enc=encryptedRaw.hex())
    e.match('only one field from raw, enc, hash is expected')



from indy_common.constants import SCHEMA_NAME, SCHEMA_VERSION, SCHEMA_ATTR_NAMES, SCHEMA_FROM, \
    CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_TAG, CLAIM_DEF_PUBLIC_KEYS, CLAIM_DEF_FROM, \
    CLAIM_DEF_TAG_DEFAULT, CLAIM_DEF_CL

from plenum.common.constants import DATA
from plenum.common.request import Request
from plenum.common.txn_util import get_payload_data


# TODO: use data classes instead


# SCHEMA

def get_write_schema_name(req):
    return req.operation[DATA][SCHEMA_NAME]


def get_txn_schema_name(txn):
    return get_payload_data(txn)[DATA][SCHEMA_NAME]


def get_write_schema_version(req: Request):
    return req.operation[DATA][SCHEMA_VERSION]


def get_txn_schema_version(txn):
    return get_payload_data(txn)[DATA][SCHEMA_VERSION]


def get_write_schema_attr_names(req: Request):
    return req.operation[DATA][SCHEMA_ATTR_NAMES]


def get_txn_schema_attr_names(txn):
    return get_payload_data(txn)[DATA][SCHEMA_ATTR_NAMES]


def get_read_schema_name(req: Request):
    return req.operation[DATA][SCHEMA_NAME]


def get_read_schema_version(req: Request):
    return req.operation[DATA][SCHEMA_VERSION]


def get_read_schema_from(req: Request):
    return req.operation[SCHEMA_FROM]


def get_reply_schema_name(reply):
    return reply[DATA][SCHEMA_NAME]


def get_reply_schema_version(reply):
    return reply[DATA][SCHEMA_VERSION]


def get_reply_schema_attr_names(reply):
    return reply[DATA].get(SCHEMA_ATTR_NAMES)


def get_reply_schema_from(reply):
    return reply[SCHEMA_FROM]


# CLAIM DEF

def get_write_claim_def_signature_type(req: Request):
    return req.operation.get(CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_CL)


def get_write_claim_def_schema_ref(req: Request):
    return req.operation[CLAIM_DEF_SCHEMA_REF]


def get_write_claim_def_tag(req: Request):
    return req.operation.get(CLAIM_DEF_TAG, CLAIM_DEF_TAG_DEFAULT)


def get_write_claim_def_public_keys(req: Request):
    return req.operation[CLAIM_DEF_PUBLIC_KEYS]


def get_txn_claim_def_signature_type(txn):
    return get_payload_data(txn).get(CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_CL)


def get_txn_claim_def_schema_ref(txn):
    return get_payload_data(txn)[CLAIM_DEF_SCHEMA_REF]


def get_txn_claim_def_tag(txn):
    return get_payload_data(txn).get(CLAIM_DEF_TAG, CLAIM_DEF_TAG_DEFAULT)


def get_txn_claim_def_public_keys(txn):
    return get_payload_data(txn)[CLAIM_DEF_PUBLIC_KEYS]


def get_read_claim_def_signature_type(req: Request):
    return req.operation.get(CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_CL)


def get_read_claim_def_schema_ref(req: Request):
    return req.operation[CLAIM_DEF_SCHEMA_REF]


def get_read_claim_def_tag(req: Request):
    return req.operation.get(CLAIM_DEF_TAG, CLAIM_DEF_TAG_DEFAULT)


def get_read_claim_def_from(req: Request):
    return req.operation[CLAIM_DEF_FROM]



import pytest

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.authorize import auth_map
from indy_common.constants import NYM, ROLE, ENDORSER
from indy_node.test.auth_rule.auth_framework.basic import roles_to_string, AuthTest
from indy_node.test.auth_rule.helper import create_verkey_did
from plenum.common.exceptions import RequestRejectedException

from indy_node.test.helper import build_auth_rule_request_json


class AddNewRoleTest(AuthTest):
    def __init__(self, action_id: str, creator_wallet, env):
        super().__init__(env, action_id)
        self.role = self.action.new_value
        self.role_string = roles_to_string[self.role]
        self.creator_wallet = creator_wallet
        self.checker_wallet = None

    def prepare(self):
        self.phase_req_1 = self.get_nym()
        self.phase_req_2 = self.get_nym()
        self.phase_req_3 = self.get_nym()

        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        self.nym_for_new_rule = self._get_nym_for_new_rule()

    def run(self):
        # Step 1. Check default auth rule
        self.send_and_check(self.phase_req_1, wallet=self.creator_wallet)

        # Step 2. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 3. Check, that we cannot add new nym with role by old way
        with pytest.raises(RequestRejectedException):
            self.send_and_check(self.phase_req_2, wallet=self.creator_wallet)

        # Step 4. Check, that new rule is working
        self.send_and_check(self.nym_for_new_rule, wallet=self.checker_wallet)

        # Step 5. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 6. Check, that default auth rule works
        self.send_and_check(self.phase_req_3, wallet=self.creator_wallet)

    def result(self):
        pass

    def get_nym(self):
        wh, _ = self.creator_wallet
        did, verkey = create_verkey_did(self.looper, wh)
        return self._build_nym(self.creator_wallet,
                               self.role_string,
                               did,
                               verkey=verkey,
                               skipverkey=False)

    def _get_nym_for_new_rule(self):
        wh, _ = self.checker_wallet
        did, verkey = create_verkey_did(self.looper, wh)
        return self._build_nym(self.checker_wallet,
                               self.role_string,
                               did,
                               verkey=verkey,
                               skipverkey=False)

    def get_changed_auth_rule(self):
        self.checker_wallet = self.env.role_to_wallet[ENDORSER]
        constraint = AuthConstraint(role=ENDORSER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.creator_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=NYM,
            field=ROLE,
            new_value=self.role,
            constraint=constraint.as_dict
        )


class AddNewTrusteeTest(AddNewRoleTest):
    def __init__(self, env, action_id=auth_map.add_new_trustee.get_action_id()):
        super().__init__(action_id, env.sdk_wallet_trustee, env)


class AddNewStewardTest(AddNewRoleTest):
    def __init__(self, env, action_id=auth_map.add_new_steward.get_action_id()):
        super().__init__(action_id, env.sdk_wallet_trustee, env)


class AddNewEndorserTest(AddNewRoleTest):
    def __init__(self, env, action_id=auth_map.add_new_endorser.get_action_id()):
        super().__init__(action_id, env.sdk_wallet_trustee, env)


class AddNewNetworkMonitorTest(AddNewRoleTest):
    def __init__(self, env, action_id=auth_map.add_new_network_monitor.get_action_id()):
        super().__init__(action_id, env.sdk_wallet_trustee, env)


class AddNewIdentityOwnerTest(AddNewRoleTest):
    def __init__(self, env, action_id=auth_map.add_new_identity_owner.get_action_id()):
        super().__init__(action_id, env.sdk_wallet_trustee, env)



#!/usr/bin/python3.5
import sys
import traceback

from stp_core.common.log import getlogger

logger = getlogger()

ENV_FILE_PATH = "/etc/indy/indy.env"

node_ip_key = 'NODE_IP'
client_ip_key = 'NODE_CLIENT_IP'


def are_ips_present_in_env():
    node_ip_present = False
    client_ip_present = False

    with open(ENV_FILE_PATH, "r") as fenv:
        for line in fenv.readlines():
            key = line.split('=')[0].strip()
            if key == node_ip_key:
                node_ip_present = True
                logger.info("Node IP is present in '{}': {}".format(ENV_FILE_PATH, line.strip()))
            elif key == client_ip_key:
                client_ip_present = True
                logger.info("Client IP is present in '{}': {}".format(ENV_FILE_PATH, line.strip()))
    return node_ip_present, client_ip_present


def append_ips_to_env(node_ip, client_ip):
    node_ip_key = 'NODE_IP'
    client_ip_key = 'NODE_CLIENT_IP'

    if node_ip is None and client_ip is None:
        return

    with open(ENV_FILE_PATH, "a") as fenv:
        fenv.write("\n")
        if node_ip is not None:
            line = node_ip_key + "=" + node_ip
            logger.info("Appending line to '{}': {}".format(ENV_FILE_PATH, line))
            fenv.write("{}\n".format(line))
        if client_ip is not None:
            line = client_ip_key + "=" + client_ip
            logger.info("Appending line to '{}': {}".format(ENV_FILE_PATH, line))
            fenv.write("{}\n".format(line))


def migrate_all():
    node_ip = None
    client_ip = None
    node_ip_present, client_ip_present = are_ips_present_in_env()

    if not node_ip_present:
        node_ip = "0.0.0.0"

    if not client_ip_present:
        client_ip = "0.0.0.0"

    if node_ip is not None or client_ip is not None:
        try:
            append_ips_to_env(node_ip, client_ip)
        except Exception:
            logger.error(traceback.print_exc())
            logger.error("Could not append node and client IPs to indy env file")
            return False
    else:
        logger.info("No modification of env file is needed")

    return True


if migrate_all():
    logger.info("Node/client IPs migration complete")
else:
    logger.error("Node/client IPs migration failed.")
    sys.exit(1)



import json
import time

from indy import blob_storage
from indy.anoncreds import issuer_create_and_store_revoc_reg
from indy.ledger import build_revoc_reg_def_request, build_revoc_reg_entry_request

from plenum.common.types import f

from plenum.common.constants import STATE_PROOF, ROOT_HASH, MULTI_SIGNATURE, PROOF_NODES, MULTI_SIGNATURE_SIGNATURE, \
    MULTI_SIGNATURE_PARTICIPANTS, MULTI_SIGNATURE_VALUE, MULTI_SIGNATURE_VALUE_LEDGER_ID, \
    MULTI_SIGNATURE_VALUE_STATE_ROOT, MULTI_SIGNATURE_VALUE_TXN_ROOT, MULTI_SIGNATURE_VALUE_POOL_STATE_ROOT, \
    MULTI_SIGNATURE_VALUE_TIMESTAMP, TYPE, DATA, TXN_TIME, TXN_PAYLOAD, TXN_PAYLOAD_METADATA, TXN_PAYLOAD_METADATA_FROM, \
    TXN_PAYLOAD_DATA, TXN_METADATA, TXN_METADATA_SEQ_NO, TXN_METADATA_TIME, TXN_TYPE
from indy_common.constants import VALUE, GET_REVOC_REG_DEF, GET_REVOC_REG, GET_REVOC_REG_DELTA, \
    ACCUM_TO, ISSUED, STATE_PROOF_FROM, ACCUM_FROM, CRED_DEF_ID, REVOC_TYPE, TAG, REVOC_REG_DEF_ID, FROM, TO, TIMESTAMP
from common.serializers.serialization import state_roots_serializer, proof_nodes_serializer
from plenum.common.util import randomString
from plenum.test.helper import sdk_sign_request_from_dict
from state.pruning_state import PruningState
from indy_common.state import domain


def get_cred_def_id(submitter_did, schema_id, tag):
    cred_def_marker = domain.MARKER_CLAIM_DEF
    signature_type = 'CL'
    cred_def_id = ':'.join([submitter_did, cred_def_marker, signature_type, str(schema_id), tag])
    return cred_def_id


def get_revoc_reg_def_id(author_did, revoc_req):
    return ":".join([author_did,
                     domain.MARKER_REVOC_DEF,
                     revoc_req['operation'][CRED_DEF_ID],
                     revoc_req['operation'][REVOC_TYPE],
                     revoc_req['operation'][TAG]])


def get_revoc_reg_entry_id(submitter_did, def_revoc_id):
    entry_revoc_id = ':'.join([submitter_did, domain.MARKER_REVOC_REG_ENTRY, def_revoc_id])
    return entry_revoc_id


def create_revoc_reg(looper, wallet_handle, submitter_did, tag, cred_def_id,
                     issuance):
    tails_writer_config = json.dumps({'base_dir': 'tails', 'uri_pattern': ''})
    tails_writer = looper.loop.run_until_complete(blob_storage.open_writer(
        'default', tails_writer_config))
    _, revoc_reg_def_json, revoc_reg_entry_json = \
        looper.loop.run_until_complete(issuer_create_and_store_revoc_reg(
            wallet_handle, submitter_did, "CL_ACCUM", tag, cred_def_id,
            json.dumps({"max_cred_num": 5, "issuance_type": issuance}),
            tails_writer))
    return looper.loop.run_until_complete(
        build_revoc_reg_def_request(submitter_did, revoc_reg_def_json))


def create_revoc_reg_entry(looper, wallet_handle, submitter_did, tag, cred_def_id,
                           issuance):
    tails_writer_config = json.dumps({'base_dir': 'tails', 'uri_pattern': ''})
    tails_writer = looper.loop.run_until_complete(blob_storage.open_writer(
        'default', tails_writer_config))
    revoc_reg_def_id, revoc_reg_def_json, revoc_reg_entry_json = \
        looper.loop.run_until_complete(issuer_create_and_store_revoc_reg(
            wallet_handle, submitter_did, "CL_ACCUM", tag, cred_def_id,
            json.dumps({"max_cred_num": 5, "issuance_type": issuance}),
            tails_writer))
    revoc_reg_def = looper.loop.run_until_complete(
        build_revoc_reg_def_request(submitter_did, revoc_reg_def_json))
    entry_revoc_request = looper.loop.run_until_complete(
        build_revoc_reg_entry_request(
            submitter_did, revoc_reg_def_id, "CL_ACCUM", revoc_reg_entry_json))

    return revoc_reg_def, entry_revoc_request


def prepare_for_state(result):
    request_type = result[TYPE]
    if request_type == GET_REVOC_REG_DEF:
        return domain.prepare_get_revoc_def_for_state(result)
    if request_type == GET_REVOC_REG:
        return domain.prepare_get_revoc_reg_entry_accum_for_state(result)
    if request_type == GET_REVOC_REG_DELTA:
        if ISSUED in result[DATA][VALUE]:
            return domain.prepare_get_revoc_reg_entry_for_state(result)
        else:
            return domain.prepare_get_revoc_reg_entry_accum_for_state(result)
    raise ValueError("Cannot make state key for "
                     "request of type {}"
                     .format(request_type))


def validate_proof(result):
    """
    Validates state proof
    """
    state_root_hash = result[STATE_PROOF]['root_hash']
    state_root_hash = state_roots_serializer.deserialize(state_root_hash)
    proof_nodes = result[STATE_PROOF]['proof_nodes']
    if isinstance(proof_nodes, str):
        proof_nodes = proof_nodes.encode()
    proof_nodes = proof_nodes_serializer.deserialize(proof_nodes)
    key, value = prepare_for_state(result)
    valid = PruningState.verify_state_proof(state_root_hash,
                                            key,
                                            value,
                                            proof_nodes,
                                            serialized=True)
    return valid


def check_valid_proof(reply):
    result = reply['result']
    assert STATE_PROOF in result

    state_proof = result[STATE_PROOF]
    assert ROOT_HASH in state_proof
    assert state_proof[ROOT_HASH]
    assert PROOF_NODES in state_proof
    assert state_proof[PROOF_NODES]
    assert MULTI_SIGNATURE in state_proof

    multi_sig = state_proof[MULTI_SIGNATURE]
    assert multi_sig
    assert multi_sig[MULTI_SIGNATURE_PARTICIPANTS]
    assert multi_sig[MULTI_SIGNATURE_SIGNATURE]
    assert MULTI_SIGNATURE_VALUE in multi_sig

    multi_sig_value = multi_sig[MULTI_SIGNATURE_VALUE]
    assert MULTI_SIGNATURE_VALUE_LEDGER_ID in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_LEDGER_ID]
    assert MULTI_SIGNATURE_VALUE_STATE_ROOT in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_STATE_ROOT]
    assert MULTI_SIGNATURE_VALUE_TXN_ROOT in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_TXN_ROOT]
    assert MULTI_SIGNATURE_VALUE_POOL_STATE_ROOT in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_POOL_STATE_ROOT]
    assert MULTI_SIGNATURE_VALUE_TIMESTAMP in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_TIMESTAMP]
    if result[TYPE] == GET_REVOC_REG_DELTA:
        if STATE_PROOF_FROM in result[DATA] and result[DATA][STATE_PROOF_FROM]:
            reply_from = {DATA: result[DATA][VALUE][ACCUM_FROM],
                          STATE_PROOF: result[DATA][STATE_PROOF_FROM],
                          TYPE: result[TYPE],
                          f.SEQ_NO.nm: result[DATA][VALUE][ACCUM_FROM][f.SEQ_NO.nm],
                          TXN_TIME: result[DATA][VALUE][ACCUM_FROM][TXN_TIME]
                          }
            assert validate_proof(reply_from)
        reply_to = {DATA: result['data'][VALUE][ACCUM_TO],
                    STATE_PROOF: result[STATE_PROOF],
                    TYPE: result[TYPE],
                    f.SEQ_NO.nm: result[f.SEQ_NO.nm],
                    TXN_TIME: result[TXN_TIME]
                    }
        assert validate_proof(reply_to)
    else:
        assert validate_proof(result)

def build_get_revoc_reg_delta(looper,
                              sdk_wallet_steward):
    data = {
        REVOC_REG_DEF_ID: randomString(10),
        TXN_TYPE: GET_REVOC_REG_DELTA,
        FROM: 10,
        TO: 20,
    }
    revoc_reg_delta_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return revoc_reg_delta_req

def build_get_revoc_reg_entry(looper,
                              sdk_wallet_steward):
    data = {
        REVOC_REG_DEF_ID: randomString(10),
        TXN_TYPE: GET_REVOC_REG,
        TIMESTAMP: int(time.time())
    }
    revoc_reg_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return revoc_reg_req
